"""
Main class of the asciidata module

@author: Martin Kuemmel, Jonas Haase
@organization: Space Telescope - European Coordinating Facility (ST-ECF)
@license: Gnu Public Licence
@contact: mkuemmel@eso.org
@version: 0.9
@since: 2005/09/13
"""

import string, sys, os, types
from asciiheader import *
from asciicolumn import *
from asciierror import *
from asciiutils import *

class NullData(object):
    """
    Null class as a parent class for the AsciiData class

    This parent classs of the AsciiData class offers to create
    a new AsciiData instance without a file to read from.
    All elements are set to None, but of course can later
    be filled by the user.
    """
    def __init__(self, ncols, nrows, null=None):
        """
        Constructor for the NullData Class

        Creates an empty AsciiData instance with columns and
        rows as specified. All entries are 'None'.

        @param ncols: the number of columns to be created
        @type ncols: integer
        @param nrows: the number of rows to be created
        @type nrows: integer
        @param null: string to be interpretet as NULL
        @type null: string
        """
        # set the default null string
        if null:
            self._null = [string.strip(null)]
        else:
            self._null  = ['Null']

        # create the colum list
        self.columns = []
        for index in range(ncols):
            # get the column name
            colname = self._def_colname(index)
            # create and append an empty column
            self.columns.append(AsciiColumn(nrows=nrows, colname=colname,
                                            null=self._null))
    def _def_colname(self, index):
        """
        Gives the default column name.

        The method composes and returns the
        default column name for a column at a
        given index.
        
        @param index: the index of the column
        @type index: integer
        """
        return 'column'+str(index+1)

class AsciiData(NullData):
    """
    Basic class in the AstroAsciiData project

    This class and its methods forms the complete API for the
    for the
    """
    def __init__(self, filename=None, ncols=0, nrows=0, null=None,
                 delimiter=None, comment=None):
        """
        Constructor for the AsciiData Class

        The data is taken from  a file specified in the input.
        As addition, a NULL string, a delimiter string and a comment
        string can be specified. The ascii data is read in from the
        file and stored in a list of Columns

        @param filename: the filename to create the AsciiData from
        @type filename: string
        @param ncols: the number of columns to be created
        @type ncols: integer
        @param nrows: the number of rows to be created
        @type nrows: integer
        @param null: string to be interpretet as NULL
        @type null: string
        @param delimiter: string to be used as delimiter
        @type delimiter: string
        @param comment: string to be used as comment
        @type comment: string
        """
        self.ncols = 0
        self.nrows = 0

        # set the default comment
        if comment:
            self._comment = comment
        else:
            self._comment = '#'
            
        # set the default null string
        if null:
            self._null = [string.strip(null)]
        else:
            self._null  = ['Null', 'NULL', 'None', '*']

        # set the delimiter
        self._delimiter = delimiter

        # set the separator
        self._separator = Separator(delimiter)

        # create the header
        self.header = Header(filename, self._comment)

        # check whether a filename is given
        if filename != None:
            # check whether the file exists
            if os.path.exists(filename):
                self._filename = filename
            else:
                raise "Filename: " + filename + " does not exist!" 

            # load in all data from the files
            self.columns = self._load_columns(filename, self._null,
                                              self._comment, self._separator)
        else:
            # set the filename to none
            self._filename = None

            # check whether valid numbers where given
            if nrows > 0 and ncols > 0:
                # create the empty instance
                super(AsciiData, self).__init__(ncols, nrows, null)
            else:
                raise "Number of columns, rows: " \
                      + str(ncols) + str(nrows)  + " are not reasonable!"

        # find the number of undefined columns
        self._undef_cols = self._find_undefined_cols(self.columns)

        # find the number of columns and rows
        self.ncols = len(self.columns)
        if self.ncols:
            self.nrows = self.columns[0].get_nrows()
                
    def __getitem__(self, element):
        """
        Defines the list operator for indexing

        This method returns the value at a given index.
        In the current class this means the method returns
        the column at the given index.

        @param element: either column index or name
        @type element: string/integer

        @return: a column
        @rtype: AsciiColumn
        """

        try:
            index = self._loc_column(element)
        except ColumnError:
            index = self.append(element)

        # return the desired column
        return self.columns[index]

    def __setitem__(self, element, column):
        """
        Defines the list operator for indexed assignement

        The method inserts a column to the class at the
        specified index. As of now, it is not possible
        to create extra columns with this method,
        only existing columns can be overwritten.

        @param element: either column index or name
        @type element: string/integer
        @param column: the column to assign to an index
        @type column: AsciiColumn
        """

        index = self._loc_column(element)

        # check whether the column does have the same number
        # of rows as the class
        # raise an error if not
        if column.get_nrows() != self.nrows:
            raise 'Nrows: '+str(column.get_nrows())+' different than nrows: '\
                  +str(self.nrows)+'!!'

        # check whether the column has a name
        if not column.colname:
            # give it a default name
            column.colname = self._def_colname(index)

        # assign the new column
#        for ii in range(column.get_nrows()):
        self.columns[index] = column.copy()

    def __delitem__(self, element):
        """
        Deletes an index.

        The method deletes a column specified in the input.
        The column can be specified either by the column
        name or the index.

        @param element: either column index or name
        @type element: string/integer
        """
        # get the index from the input
        index = self._loc_column(element)

        # delete the column
        del self.columns[index]

        # adjust the number of columns
        self.ncols -= 1

    def __str__(self):
        """
        Defines a string method for the object.

        Gives a simple string method such that str(AsciiData)
        does work. The formatting is close to the formatting
        for the output to files.

        @return: the string representation of the object
        @rtype: string
        """
        bigstring = ''
        
        # take the object delimiter or ' '
        if not self._delimiter:
            delim = ' '
        else:
            delim = ' ' + self._delimiter + ' '

        # add the header to the string
        bigstring = bigstring + str(self.header)
        
        # go over each row
        for ii in range(self.nrows):
            # create the string list
            strlist = self._row_tostring(ii)

            # treat the first line different
            if ii:
                # transform the listing to one string and append it
                # put a linefeed at the beginning
                bigstring = bigstring + '\n' + delim.join(strlist)
            else:
                # transform the listing to one string and append it
                bigstring = bigstring + delim.join(strlist)

        # return the string
        return bigstring

    def flush(self):
        """
        Prints the current status to the file.

        The methods gives the opportunity to replace the data in
        the AsciiData with the current version in memory.
        """
        if self._filename != None:
            # well, that an easy job
            self.writeto(self._filename)
        else:
            raise "No filename given. Use 'writeto()' instead."

    def writeto(self, filename):
        """
        Prints the AsciiData to a new file.

        The method prints the current status of the
        object to a new file. The name of the file
        is given in the input. An already existing 
        file is replaced.

        @param filename: the filename to write the object to
        @type filename: string
        """
        # open the file
        fstream = file(filename,'w+')

        # open a printstream
        nprinter = NicePrinter(fstream, delimiter=self._delimiter)

        # print everything to the stream
        self._print_tostream(nprinter)

        #close the file
        fstream.close()

        # use the given name as class filename
        # if no one is yet defined
        if self._filename == None:
            self._filename = filename

    def info(self):
        """
        Print class info to the screen.

        The method gives some basic information on the
        class. The output is directly written onto
        the screen.

        @return: the string representing the information
        @rtype: string
        """

        # define the return string
        bigstring = ''

        # assemble the basic table information
        bigstring += 'File:       ' + str(self._filename) +'\n'
        bigstring += 'Ncols:      ' + str(self.ncols) + '\n'
        bigstring += 'Nrows:      ' + str(self.nrows) + '\n'
        bigstring += 'Delimiter:  ' + str(self._delimiter) + '\n'
        bigstring += 'Null value: ' + str(self._null) + '\n'
        bigstring += 'Comment:    ' + str(self._comment) + '\n'
        
        # go over each column and add
        # the individual column info
        for col in self.columns:
            bigstring += col.info()

        # return the result
        return bigstring

    def append(self, colname):
        """
        Appends a new column to the object.

        This method creates and appends a new column to the 
        object. The new column must be specified with a name.
        The new column doe have only Null entries.

        @param colname: the name of the column
        @type colname: string
        """
        # check whether the column name does exist
        # raise a warning if yes
        if self.find(colname) > -1:
            raise 'Column with name: '+colname+' does just exist!'      

        # get the index of the new column
        index = self.ncols

        # create and append the new column
        self.columns.append(AsciiColumn(nrows=self.nrows, colname=colname,
                                        null=self._null))
        
        # adjust the number of columns
        self.ncols +=1

        #return the index of the column
        return index

    def find(self, colname):
        """
        Finds the column number for a name.

        The method looks through all columns of the instance
        for a matching column name. In case the column name exists,
        the column index is returned. If the column name does
        not exist, -1 is returned.

        @param colname: the name of the column
        @type colname: string

        @return: the index of the column, or -1
        @rtype: integer
        """
        for index in range(len(self.columns)):
            if self.columns[index].colname == colname:
                return index
        return -1

    def delete(self, start, end=None):
        """
        Deletes a row slice or element from all columns.

        The method deletes one or several rows from all columns.
        It uses the __delelte__ or __delitem__ operators
        in the AsciiColumn class.       

        @param start: the starting row index
        @type start: integer
        @param end: the end row index
        @type end: integer
        """
        if end:
            if start < self.nrows:
                # go over each column
                for col in self.columns:
                    # delete the row
                    del col[start: end]

                # adjust the number of rows
                self.nrows -= end-start

        else:
            # go over each column
            for col in self.columns:
                # delete the row
                del col[start]
            
            # adjust the number of rows
            self.nrows -= 1

        # make a lower limit to the number of rows
        if self.nrows < 0:
            self.nrows = 0

    def newcomment(self, comment):
        """
        Define a new comment string

        @param comment: the new null string
        @type comment: string
        """
        # store the new null element
        self._comment = comment
        self.header.set_comment(comment)

    def newnull(self, newnull):
        """
        Define a new null string

        @param newnull: the new null string
        @type newnull: string
        """
        # store the new null element
        self._null[0] = newnull

        # store the new null in the columns
        for column in self.columns:
            column._null = newnull

    def newdelimiter(self, delimiter):
        """
        Set a new delimiter string
        
        @param delimiter: the new delimiter string
        @type delimiter: string
        """
        # set the new delimiter
        self._delimiter = delimiter

        # set the separator
        self._separator = Separator(delimiter)
 
    def insert(self, nrows, start=0):
        """
        Inserts one or several rows

        The method inserts one or several rows into all
        columns of the class instance. The number of rows 
        as well as the positioning of the new rows are
        specified in the input. The parameter 'start'
        gives the index which the first inserted row
        will have.
        Setting "start=-1" means appending the rows at
        the end of the columns

        @param nrows: the number of rows to add
        @type nrows: integer
        @param start: the position of the inserted rows 
        @type start: integer
        """

        # go over all columns
        for col in self.columns:
            # add none elements at the end
            for ii in range(nrows):
                col.add_element(None)

        # check whether the new rows are inserted inside
        # the old rows, then the elements must be moved
        if start < self.nrows and start != -1:

            # go over all columns
            for col in self.columns:

                # repeat over rows to be inserted
                for ii in range(self.nrows-start):
                    # reorder the column elements
                    index = self.nrows - ii - 1
                    col[index+nrows] = col[index]

                # repeat over rows to be inserted
                for ii in range(nrows):
                    # insert None in the new rows
                    index = ii + start
                    col[index] = None


        # update the number of rows
        self.nrows = self.columns[0].get_nrows()

    def _loc_column(self, element):
        """
        Localizes a column

        The method localizes the column from any possible input.
        Possible input is either the column name or column index.
        Basic checks are done whether the column exists.

        @param element: either column index or name
        @type element: string/integer

        @return: the column index 
        @rtype: integer
        """
        # create an element
        elem = Element(element)

        # check the types and derive the column index
        if elem.get_type() == types.IntType:
            # check for -1, which indicates the last column
            if element == -1:
                # set the index of the last column
                index = self.ncols-1
            else:
                # set the index to the input index
                index = element
        elif elem.get_type() == types.StringType:
            index = self.find(element)

        # check whether the column index exists
        # raise an error if not
        if index > self.ncols-1:
            raise 'Index: '+str(index)+' is larger than ncols: '\
                  +str(self.ncols)+'!!'
        elif index < 0:
            raise ColumnError('Column name: "'+element+'" does not exist!')

        # return the index
        return index

    def _load_columns(self, filename, null, comment, separator):
        """
        Transforms the content of a file into columns

        Opens the file, defines the columns, adds all data rows,
        and returns the columns.
        
        @param filename: the filename to create the AsciiData from
        @type filename: string
        @param null: string to be interpreted as NULL
        @type null: string
        @param separator: string to be used as delimiter
        @type separator: string
        @param comment: string to be used as comment
        @type comment: string

        @return: the columns loaded
        @rtype: [AsciiColumn]
        """

        undef_cols = []
        collist    = []

        lnum = 0
        # open the file, and parse through all rows
        for line in file(filename, 'r'):

#           print lnum, line
            # throw away trailing and leading whitespaces
            str_line = string.strip(line)
            if len(str_line) < 1 or str_line[0] == comment:
                continue

            # if collumns exist, add a row 
            if collist:
                self._add_row(collist, line,  null, separator)
            # if columns do not exist, define them
            else:
                collist    = self._define_cols(line,  null, separator)
            
            lnum +=1
                
        # return the column list
        return collist

    def _find_undefined_cols(self, collist):
        """
        Finds undefined columns

        The method finds undefined columns in a column list.
        An undefined column is a column with the flag "self._defined"
        not set. This means that column type and column format
        are not specified, and the column elements are Null.
        The indices of the undefined columns is returned as a list
        
        @param collist: the list of existing columns
        @type collist: list of AsciiColumns

        @return: a list with the indices of undefined columns
        @rtype: [integer]
        """
        undefined = []

        # go over each column
        index=0
        for col in collist:
            # check whether the column is defined
            # append the index to the list if not
            if not col.get_defined():
                undefined.append(index)
            # increment the index
            index = index+1

        # return the list
        return undefined

    def _add_row(self, collist, line,  null, separator):
        """
        Adds a line from the file to the column list.

        The method gets a line from the input file.
        The line is split up into its items.
        Then each item is added to the column 
        it belongs to. Items matching the NULL
        string are added as "None". A delimiter
        is taken into account in the splitting,
        if specified.

        @param collist: the list of existing columns
        @type collist: list of AsciiColumns
        @param line: the line to be added to the columns
        @type line: string
        @param null: string to be interpretet as NULL
        @type null: string
        @param separator: string to be used as delimiter
        @type separator: string
        """
        # split the line, either according toa whitespace,
        # or according to a specified delimiter
        items = separator.separate(line)

        # check whether there is an item for each column
        if len(collist) != len(items):
            raise "Number of columns does not fit to number of items in " + line

        # go over each item
        index = 0
        for item in items:
            
            # check whether the item is NULL.
            # add the item to the column,
            # using 'None' for NULL items
            if null.count(string.strip(item)) > 0:
                collist[index].add_element(None)
            else:
                collist[index].add_element(item)

            # increment the index
            index += 1

    def _define_cols(self, line,  null, separator):
        """
        Defines the columns from an input line.

        The method splits an ascii line from the input file into its
        items. For each item a new column is created and added
        to a column list. The column list is finally returned.
        
        @param line: the line to be added to the columns
        @type line: string
        @param null: string to be interpretet as NULL
        @type null: string
        @param separator: string to be used as delimiter
        @type separator: string

        @return: the columns created
        @rtype: [AsciiColumn]
        """
        collist = []
        
        # split the line, either according toa whitespace,
        # or according to a specified delimiter
        items = separator.separate(line)

        # go over each item, and create a column
        # for each. NULL items are transformed to 'None'
        index = 0
        for item in items:
            # make the default column name
            colname = self._def_colname(index)
            if null.count(string.strip(item)) > 0:
                collist.append(AsciiColumn(element=[None], colname=colname,
                                           null=null))
            else:
                collist.append(AsciiColumn(element=[item], colname=colname,
                                           null=null))

            # increment the index
            index += 1
            
        # return the column list
        return collist

    def _print_tostream(self, nprinter):
        """
        Prints the AsciiData to a stream

        The method forms for each row in the AsciiData a list
        with formated strings, each list element representing
        one element. The list is sent to a printing stream
        which is responsible for the output.

        @param nprinter: the NicePrinter object with the stream
        @type nprinter: NicePrinter     
        """
        # print the header
        nprinter.print_string(str(self.header))

        # print the data
        # go over each row
        for ii in range(self.nrows):
            # create the string list
            strlist = self._row_tostring(ii)

            # send the list to the printer
            nprinter.print_list(strlist)

    def _row_tostring(self, index):
        """
        Creates the formatted string list for one row.

        The method extracts from each column the formatted
        string representation of the element in a specified
        row. The list of strings is returned.

        @param index:
        @type index: integer
        
        @return: the list with formatted strings
        @rtype: [string]
        """
        # initialize the list
        strlist = []

        # go over each column
        for jj in range(self.ncols):
            # append the string of the requested
            # element to the list
            strlist.append(self.columns[jj].fprint_elem(index))

        # return the list
        return strlist

