"""
Various header classes to be part of the asciidata class 

@author: Martin Kuemmel, Jonas Haase
@organization: Space Telescope - European Coordinating Facility (ST-ECF)
@license: Gnu Public Licence
@contact: mkuemmel@eso.org
@version: 0.9
@since: 2005/09/13
"""

import string

class Header(object):
    """
    The header object

    This object offers the possibility to store additional
    information such as change comments or column information.
    This additional information may just be present at the
    beginning of the data file or later be added.
    """
    def __init__(self, filename=None, comment=None):
        """
        Constructor for the Header class

        @param filename: the data file
        @type filename: string
        @param comment: the comment string
        @type comment: string
        """
        # store the comment
        self._comment = comment

        # retrieve the comment from the data file
        if filename == None:
            self.hdata = []
        else:
            self.hdata = self._load_header(filename, comment)

    def __str__(self):
        """
        Defines a string method for the object

        @return: the string representation
        @rtype: string
        """

        # start the string
        hstring = ''

        # add the different items
        for line in self.hdata:
            if len(line) > 0:
                hstring += self._comment + ' ' + line + '\n'
            else:
                hstring += self._comment + '\n'

        # return the string
        return hstring

    def append(self, hlist):
        """
        Append something to the header data
    
        @param hlist: the string to append
        @type hlist: string
        """
        # split the string to lines
        hitems = string.split(hlist,'\n')

        # for each line
        for item in hitems:
            # append the new content
            # to the header content
            self.hdata.append(string.strip(item))

    def _make_hdata(self, item):
        """
        Forms a new header item
    
        @param item: the item
        @type item: string
        @return: the proper string
        @rtype: string
        """
        # add the comment and a newline
        newdata = self._comment + ' ' + item + '\n'

        # return the string
        return newdata

    def _load_header(self, filename, comment):
        """
        Loads the header from the data file

        @param filename: the data file
        @type filename: string
        @param comment: the comment string
        @type comment: string
        """
        # start the item list
        data = []
        
        # open the data file and go over its rows
        for line in file(filename, 'r'):
            # throw away trailing and leading whitespaces
            str_line = string.strip(line)

            # append a comment to the list
            if str_line.find(comment) == 0:
                data.append(string.strip(line[len(comment)+1:]))
            else:
                # leave the file at the first
                # non-comment line
                break
            
        # return the header list
        return data
 
    def set_comment(self, comment):
        """
        Set the comment string
        
        @param comment: the new comment string
        @type comment: string
        """
        self._comment = comment
