"""
$Revision: 1.3 $ $Date: 2010-08-25 06:47:17 $
Author: Martin Kuemmel
Affiliation: Space Telescope - European Coordinating Facility
"""
__author__ = "Martin Kuemmel <mkuemmel@eso.org>"
__date__ = "$Date: 2010-08-25 06:47:17 $"
__version__ = "$Revision: 1.3 $"
__credits__ = """This software was developed by the ACS group of the Space Telescope -
European Coordinating Facility (ST-ECF). The ST-ECF is a department jointly
run by the European Space Agency and the European Southern Observatory.
It is located at the ESO headquarters at Garching near Munich. The ST-ECF
staff supports the European astronomical community in exploiting the research
opportunities provided by the earth-orbiting Hubble Space Telescope.
"""

import sys, os, os.path, string, math
import aXe2html.cathtml.textcell, aXe2html.cathtml.axecell
import aXe2html.cathtml.imagecell, aXe2html.cathtml.htmlcat
import aXe2html.sexcat.sextractcat, aXe2html.sexcat.tablecolumn
import aXe2html.utils.namings

#--------------------------------------------------------------------
#
#--------------------------------------------------------------------
class PageMaker:
    """
    Workhorse class which is instanciated only one. That instance
    is responsible to create all the html pages. That means.
    To determine the number of content pages + creating them +
    creating the overview/index/info pages if requested.
    """
    def __init__(self, settings):
        """
        Input:
            settings - the dictionary with all available settings

        Return:
            -

        Description:
            Stores the settings which are relevant for the
            creation of the gtml into instance variables.
            The various methods then can access the settings
            via that variables.
            At the end the whole setting is stored as an instance
            variable, too. That is to pass all available
            information to crete the info page. On that page
            settings which might not be essential for the
            html-creation are displayed.
        """
        self.sextract_cat = settings["SEXTRCAT"]
        self.page_root    = settings["PAGENAME"]
        self.start_object = settings["STARTOBJ"]
        self.nobjects     = settings["NUMBEOBJ"]
        self.ptitle_root  = settings["PAGEHEAD"]
        self.phead_root   = settings["PAGETITL"]
        self.pagesize     = settings["PAGESIZE"]
        self.sortcolumn   = settings["SORTCOLU"]
        self.errorbar     = settings["ERRORBAR"]
        self.overview     = settings["OVERVIEW"]
        self.marksource   = settings["MARKSOUR"]
        self.sortorder    = settings["ASCENDIN"]
        self.outpath      = settings['OUTPPATH']
        self.imgpath      = settings['IMAGPATH']
        self.dataoutp     = settings['DATAOUTP']
        self.datapath     = settings['DATAPATH']
        self.mag_wave     = settings['MAGLAMBD']
        self.css_file     = settings['CSSFILEU']
        self.special      = 'POINT_SOURCE'
        self.page_name    = ''
        self.page_head    = ''
        self.page_title   = ''
        self.allsetts = settings

    def makepages(self):
        """
        Input:
            -

        Return:
            error code in case of problems with the catlog,
              0 otherwise

        Description:
            The central method which creates all the webpages +
            the overview/info/index pages.
            The method opens the sextractor catalog and sorts the
            objects therein.
            Then it opens the object pages, extracts the catalog
            data and creates the rows of the objects table.
            At the end the overview/info/index pages are made,
            in case that they were requested.
        """
        #--------------------------------------------------------------------
        #
        #   Open the Sextracor catalog and the direct image.
        #   Sort the catalog if requested
        sex_cat = aXe2html.sexcat.sextractcat.SexCat(self.sextract_cat)
        success = sex_cat.setaxecols(self.mag_wave, self.special)
        if success > 0:
            return 100
        if self.sortcolumn != None:
            if self.sortorder == 'YES':
                success = sex_cat.setorder(self.sortcolumn, 0)
            else:
                success = sex_cat.setorder(self.sortcolumn, 1)

            if success > 0:
                return 101
        #
        #--------------------------------------------------------------------

        decimage = aXe2html.utils.namings.DecorationFiles()
        decimage.getcssfile(self.outpath, self.css_file)


        #--------------------------------------------------------------------
        #
        #   Determine the number of objects and derive
        #   the number of html-pages which are to be created
        #
        allobjects = sex_cat.getnrows()
        if self.nobjects == -1:
            self.nobjects   = allobjects
        elif self.nobjects > allobjects:
            print ''
            print 'Sorry, I can give you only ', allobjects, ' objects.'
            print 'The catalog is not large enough!'
            print ''
            self.nobjects = allobjects

        npages = self.getnpages(self.nobjects, self.pagesize)

        #--------------------------------------------------------------------
        #
        #   Make a big loop over all objects.
        #   Do everything within that big loop
        #
        actpage    = 0
        axeobjects = []
        allaxeobs  = []
        for oneobject in range(self.nobjects):


            #--------------------------------------------------------------------
            #
            #    See whether a new page must be created.
            #    Open it, make the links on top and start the table
            if math.fmod(oneobject, self.pagesize) == 0:
                if len(axeobjects) > 0:
                    htmlname = self.makeobjectpage(actpage, axeobjects)
                    fullname = os.path.join(self.outpath, htmlname)
                    os.rename(htmlname, fullname)
                    print 'Finished page ', actpage
                    print ''
                    axeobjects = []
                actpage = actpage + 1
                if npages == 1:
                    self.page_name = self.page_root + '.html'
                    self.page_head = self.phead_root
                    self.page_title = self.ptitle_root
                else:
                    self.page_name = self.page_root + '_p' + str(actpage) + '.html'
                    self.page_head = self.phead_root + ' -- page ' + str(actpage)
                    self.page_title = self.ptitle_root + ' -- page ' + str(actpage)

            #--------------------------------------------------------------------
            #
            #   Get the sextractor information on
            #   the current object
            actobject = oneobject + self.start_object - 1
            if self.sortcolumn != None:
                axedata = sex_cat.getaxesortdata(actobject)
                axeobjects.append(sex_cat.getaxesortdata(actobject))
                allaxeobs.append(sex_cat.getaxesortdata(actobject))
            else:
                axedata = sex_cat.getaxedata(actobject)
                axeobjects.append(sex_cat.getaxedata(actobject))
                allaxeobs.append(sex_cat.getaxedata(actobject))
            #--------------------------------------------------------------------



        htmlname = self.makeobjectpage(actpage, axeobjects)
        fullname = os.path.join(self.outpath, htmlname)
        os.rename(htmlname, fullname)
        print 'Finished page ', actpage
        print ''
        if self.overview != 'NO':
            self.page_name = self.page_root + '_over.html'
            self.page_head = self.phead_root
            self.page_title = self.ptitle_root
            htmlname =  self.makeoverview(0, allaxeobs)
            fullname = os.path.join(self.outpath, htmlname)
            os.rename(htmlname, fullname)

            self.page_name  = self.page_root + '_info.html'
            self.page_head  = self.phead_root
            self.page_title = self.ptitle_root
            htmlname =  self.makeinfoview(self.allsetts, sex_cat.magwave)
            fullname = os.path.join(self.outpath, htmlname)
            os.rename(htmlname, fullname)

            self.page_name  = self.page_root + '_index.html'
            self.page_head  = self.phead_root
            self.page_title = self.ptitle_root
            htmlname =  self.makeindex(allaxeobs)
            fullname = os.path.join(self.outpath, htmlname)
            os.rename(htmlname, fullname)
            print 'Finished overview/info/index pages'
            print ''

        return 0


    def getnpages(self, nobjects, pagesize):
        """
        Input:
            nobjects - the total number of objects
            pagesize - the number of objects per page

        Return:
            npages - the number of required pages

        Description:
            Derives the number of object pages required to
            store 'nobjects' sources at 'pagesize' sources
            pper page.
        """
        if math.fmod(nobjects, pagesize) == 0:
            npages     = nobjects/pagesize
        else:
            npages     = nobjects/pagesize + 1

        return npages


    def getactlink(self, index, pagesize, objnum):
        """
        Input:
            index    - the index number of the object
                       withinin the given sorting/sorting order
            pagesize - the number of objects per page
            objnum   - the object number in the sextractor catalog

        Return:
            actstring - the linkstring to that object

        Description:
            Creates the strings for the link anchors associated to
            an object. The page where the object is located can be
            derived from'index' and 'pagesize'. the actual name
            then follows the convention 'link'+objnum, which
            is unique within a data set.
        """
        npages = self.getnpages(self.nobjects, self.pagesize)

        if math.fmod(index, pagesize) == 0:
            actpage     = index/pagesize
        else:
            actpage     = index/pagesize + 1

        if npages == 1:
            actstring = self.page_root + '.html' + '#link' + str(objnum)
        else:
            actstring = self.page_root + '_p' + str(actpage) + \
                        '.html' + '#link' + str(objnum)

        return actstring


    def makeobjectpage(self, actpage, axeobjects):
        """
        Input:
            actpage   - the number of the actual page
            axeobject - a list with axe_objects to be included
                        in the page
        Return:
            self.page_name - the name of the page which was created

        Description:
            Creates a 'usual' objectpage. That means creates an
            according htmlcat-object and filles it with header,
            links, the table and the tailer.
        """

        anchor = 'YES'

        cat_page = aXe2html.cathtml.htmlcat.HTMLCat(self.page_name,
                                                    self.page_root, actpage,
                                                    self.pagesize)
        cat_page.writehead(self.page_head, self.outpath, self.css_file)

        cat_page.makelinks(self.nobjects, self.overview)

        cat_page.writetitle(self.page_title)
        cat_page.writeobjnums(actpage)
        cat_page.opentab('axe')

        index = 0
        sequence = (actpage-1) * self.pagesize
        for axedata in axeobjects:
            index = index + 1
            sequence = sequence + 1
            cells = []

            cellname = str(sequence)
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname, 'axe'))
            cells[0].setcolhead('Seq. no.')
            cells[0].setheight('240')


            cellname = str(axedata['NUMBER'])
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname, 'axe'))
            cells[1].setcolhead('Obj. no.')
            cells[1].setheight('240')

            # create the link to the data file
            if self.dataoutp == 'YES':
                namobj   = aXe2html.utils.namings.SpectDataName(axedata['NUMBER'], 'TXT')
                fullname = os.path.join(self.outpath, self.datapath, namobj.getdefdataname())
                if os.path.isfile(fullname):
                    cells[1].setlink(fullname)

            cellname = str(axedata['MAG_AUTO'])
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname, 'axe'))
            if axedata['MAG_WAVE'] != 0:
                cells[2].setcolhead('Magnitude<br />mag('+str(axedata['MAG_WAVE']) + ')')
            else:
                cells[2].setcolhead('Magnitude')

            cells.append(aXe2html.cathtml.axecell.AxeCell(axedata, 'axe'))
            cells[3].setcolhead('Image-coordinates')

            stdimname = aXe2html.utils.namings.DirectStampName(axedata['NUMBER'], \
                                                               self.marksource)
            full_imname = os.path.join(self.imgpath, stdimname.getdefimname())

            cells.append(aXe2html.cathtml.imagecell.ImageCell(full_imname, 'axe'))
            cells[4].setcolhead('Direct image')

            stdgriname = aXe2html.utils.namings.GrismStampName(axedata['NUMBER'])
            full_griname = os.path.join(self.imgpath, stdgriname.getdefimname())

            cells.append(aXe2html.cathtml.imagecell.ImageCell(full_griname, 'axe'))
            cells[5].setcolhead('Grism stamp')


            pltfluxname = aXe2html.utils.namings.FluxPlotName(axedata['NUMBER'],self.errorbar)
            full_fluxname = os.path.join(self.imgpath, pltfluxname.getdefimname())

            pltcountname = aXe2html.utils.namings.CountsPlotName(axedata['NUMBER'],self.errorbar)
            full_countname = os.path.join(self.imgpath, pltcountname.getdefimname())

            cells.append(aXe2html.cathtml.imagecell.ImageCell(full_countname, 'axe'))
            cells[6].setcolhead('Raw spectrum')
            cells[6].setheight('240')
            cells[6].setwidth('320')
            if anchor == 'YES':
                cells[6].setanchor(str(axedata['NUMBER']))

            cells.append(aXe2html.cathtml.imagecell.ImageCell(full_fluxname, 'axe'))
            cells[7].setcolhead('Calibrated spectrum')
            cells[7].setheight('240')
            cells[7].setwidth('320')

            if index == 1:
                cat_page.gethead(cells, 35)

            cat_page.getrow(cells)

        cat_page.closetab()
        if float(self.nobjects)/float(self.pagesize) > 1.0 or self.overview == 'YES':
            cat_page.makelinkbar(self.nobjects, 'NO')
            cat_page.getlinks()
#            cat_page.getlinkbar()
        cat_page.writetail()
        cat_page.closefile()

        return self.page_name


    def makeoverview(self, actpage, axeobjects):
        """
        Input:
            actpage    - dummy parameter needed for the
                         HTMLCat-object
            axeobjects - the list with all axeobject displayed
                         on all object pages.

        Return:
            self.page_name - the name of the overview page

        Description:
            This method creates the overview page. That page serves
            as a front page to contain tables with the basic object
            information on all objects of the object pages.
            From the overview page links lead to the individual
            object on the corresponding object pages.
        """
        cat_page = aXe2html.cathtml.htmlcat.HTMLCat(self.page_name, self.page_root, actpage, \
                                           self.pagesize)
        cat_page.writehead(self.page_head, self.outpath, self.css_file)

        cat_page.makelinks(self.nobjects, 'YES')
#        cat_page.makelinkbar(self.nobjects, self.overview)

#       cat_page.writetitle(self.page_title)
        cat_page.writenicetitle(self.page_title, self.outpath, self.imgpath)
        cat_page.writesubtitle('Overview')
        cat_page.opentab()

        index = 0
        actpage = 0
        for axedata in axeobjects:
            index = index + 1
            if index == 1:
                actpage = actpage + 1
                pstring = self.page_title + ' -- page ' + str(actpage)
                cat_page.writesubsubtitle(pstring)

            cells = []

            cellname = str(index)
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setlink(self.getactlink(index, self.pagesize, axedata['NUMBER']))
            cells[0].setcolhead('Sequence number')
            cells[0].setheight('30')

            cellname = str(axedata['NUMBER'])
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setlink(self.getactlink(index, self.pagesize, axedata['NUMBER']))
            cells[1].setcolhead('Object number')

            cellname = str(axedata['MAG_AUTO'])
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            if axedata['MAG_WAVE'] != 0:
                cells[2].setcolhead('Magnitude<br />mag('+str(axedata['MAG_WAVE']) + ')')
            else:
                cells[2].setcolhead('Magnitude')

            cellname = str(axedata['X_IMAGE'])
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[3].setcolhead('X-coord')

            cellname = str(axedata['Y_IMAGE'])
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[4].setcolhead('Y-coord')

            radec = aXe2html.utils.mathlib.radectostring(axedata["X_WORLD"],axedata["Y_WORLD"])

            cellname = str(radec[0])
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[5].setcolhead('RA')
            cells[5].setheight('30')
            cells[5].setwidth('90')

            cellname = str(radec[1])
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[6].setcolhead('DEC')
            cells[6].setheight('30')
            cells[6].setwidth('90')

            if index == 1:
                cat_page.gethead(cells, 40)

            cat_page.getrow(cells)


            if math.fmod(index, self.pagesize) == 0 and index > self.pagesize-1:
                cat_page.closetab()
                cat_page.writehorizontalrule()

                actpage = actpage + 1
                pstring = self.page_title + ' -- page ' + str(actpage)
                cat_page.writesubsubtitle(pstring)
                cat_page.opentab()
                cat_page.gethead(cells, 35)


        cat_page.closetab()
        if float(self.nobjects)/float(self.pagesize) > 1.0 or self.overview == 'YES':
#               cat_page.getlinks()
            cat_page.makelinkbar(self.nobjects, self.overview)
#           cat_page.getlinkbar()
        cat_page.writetail()
        cat_page.closefile()

        return self.page_name


    def makeinfoview(self, settings, magwave):
        """
        Input:
            settings - dictionary with all settings

        Return:
            self.page_name - the name of the page

        Desription:
            Creates the info view. That info consists of a
            table were all the settings plus the corresponding
            values is stored. That is to have a protocoll
            about whats in the table at later times.
        """
        cat_page = aXe2html.cathtml.htmlcat.HTMLCat(self.page_name, self.page_root, 0, \
                                           self.pagesize)
        cat_page.writehead(self.page_head, self.outpath, self.css_file)

        cat_page.makelinks(self.nobjects, 'YES')
#        cat_page.makelinkbar(self.nobjects, self.overview)

        cat_page.writenicetitle(self.page_title, self.outpath, self.imgpath)
        cat_page.writesubtitle('Info')
        cat_page.opentab()

        cells = []
        if settings.has_key("SEXTRCAT"):
            cellname = 'Sextractor catalog:'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')
            cells[0].setwidth('250')

            cellname = settings["SEXTRCAT"]
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')
            cells[1].setheight('35')
            cells[1].setwidth('400')

            cat_page.gethead(cells, 40)
            cat_page.getrow(cells)

        cells = []
        if settings.has_key("STPIMAGE"):
            cellname = 'Grism stamp images:'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = settings["STPIMAGE"]
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings.has_key("SPCTABLE"):
            cellname = 'Table with the spectra:'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = settings["SPCTABLE"]
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings.has_key("DIRIMAGE"):
            cellname = 'Direct image:'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = settings["DIRIMAGE"]
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings.has_key("SORTCOLU"):
            cellname = 'Catalog sorted according to:'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = settings["SORTCOLU"]
            if cellname == None:
                cellname = 'not ordered'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)
        cells = []
        if magwave != 0:
            cellname = 'Displayed magnitudes at:'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = str(magwave)+' [nm]'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)
        cells = []
        if settings.has_key("ASCENDIN"):
            cellname = 'Sorting order:'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            if  settings["ASCENDIN"] == 'YES':
                cellname = 'ascending'
            elif settings["ASCENDIN"] == 'NO':
                cellname = 'descending'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings.has_key("NUMBEOBJ"):
            cellname = 'Number of objects'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            if settings["NUMBEOBJ"] == -1:
                cellname = str(self.nobjects) + ' (all)'
            else:
                cellname = str(self.nobjects)
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings.has_key("STARTOBJ"):
            cellname = 'Start at object nr.:'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = str(settings["STARTOBJ"])
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings["CSSFILEU"] != None:
            cellname = 'User supplied css file:'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = str(settings["CSSFILEU"])
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings["SELEFILE"] != None:
            cellname = 'Selection file:'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = str(settings["SELEFILE"])
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings.has_key("OUTPPATH"):
            cellname = 'Output directory'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = str(settings["OUTPPATH"])
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings.has_key("PAGEHEAD"):
            cellname = 'Header of the HTML pages'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = settings["PAGEHEAD"]
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings.has_key('PAGETITL'):
            cellname = 'Title in the HTML-headers:'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = settings["PAGETITL"]
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings.has_key("PAGESIZE"):
            cellname = 'Number of objects per html-page'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = str(settings["PAGESIZE"])
#           print cellname, ' name', self.pagesize
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings.has_key("SPECWIDT"):
            cellname = 'Length of the grism stamp images:'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = str(settings["SPECWIDT"])
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings.has_key("DIMWIDTH"):
            cellname = 'Width of the direct image stamp:'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = str(settings["DIMWIDTH"])
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings.has_key("IMAGEFOR"):
            cellname = 'Image format:'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = str(settings["IMAGEFOR"])
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)
        cells = []
        if settings.has_key("WAVUPPER") and settings.has_key("WAVLOWER"):
            cellname = 'Lower,upper wavelength range:'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = str(settings["WAVLOWER"]) + ',' + str(settings["WAVUPPER"]) + ' Angstrom'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)


        cells = []
        if settings.has_key("ERRORBAR"):
            cellname = 'Error bars plotted: '
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = settings["ERRORBAR"]
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings.has_key("MARKSOUR"):
            cellname = 'Objects marked: '
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = settings["MARKSOUR"]
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings.has_key("ONLYHTML"):
            cellname = 'Only HTML pages created:'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname = settings["ONLYHTML"]
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings.has_key("OVERVIEW"):
            cellname = 'Overview page created'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname =  settings["OVERVIEW"]
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cells = []
        if settings.has_key("DATAOUTP"):
            cellname = 'Data files created'
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[0].setcolhead('Setting')
            cells[0].setbold()
            cells[0].setheight('35')

            cellname =  settings["DATAOUTP"]
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[1].setcolhead('Value')

            cat_page.getrow(cells)

        cat_page.closetab()
#       cat_page.getlinkbar()
        cat_page.makelinkbar(self.nobjects, self.overview)
        cat_page.writetail()
        cat_page.closefile()

        return self.page_name

    def makeindex(self, axeobj):
        """
        Input:
            axeobs - all aXeobjects in the sorted order

        Return:
            self.page_name - the name of the page

        Desription:
            Creates the index page. The index page consists of
            a big table with all object number in ascending
            order. From the object numbers links lead to the
            object rows on the object pages. That page gives
            a fair and fast overview over all objects on
            all pages and provides high mobility between
            the individual objects.
        """

        nindexcols = 20

        sortcol = []
        refecol = []
        tmpcol  = []
        for oneobj in axeobj:
            sortcol.append(int(oneobj['NUMBER']))
            refecol.append(int(oneobj['NUMBER']))
            tmpcol.append(int(oneobj['NUMBER']))
        tmpcol.sort()

        for index in range(len(sortcol)):
            sortcol[index] = refecol.index(tmpcol[index])
            refecol[sortcol[index]] = None

        cat_page = aXe2html.cathtml.htmlcat.HTMLCat(self.page_name, self.page_root, 0, \
                                           self.pagesize)
        cat_page.writehead(self.page_head, self.outpath, self.css_file)

        cat_page.makelinks(self.nobjects, 'YES')
#        cat_page.makelinkbar(self.nobjects, self.overview)

#       cat_page.writetitle(self.page_title)
        cat_page.writenicetitle(self.page_title, self.outpath, self.imgpath)
        cat_page.writesubtitle('Object index')
        cat_page.opentab()

        cells = []
        index = 0
        absindex = 0
        for colmember in tmpcol:
            index = index + 1
            absindex = absindex + 1
            cellname = str(colmember)
            cells.append(aXe2html.cathtml.textcell.TextCell(cellname))
            cells[index-1].setlink(self.getactlink(sortcol[absindex-1]+1, self.pagesize,\
                                                   colmember))
            cells[index-1].setheight('25')
            cells[index-1].setwidth('40')
            if math.fmod(index, nindexcols) == 0 and index > nindexcols-1:
                cat_page.getrow(cells)
                cat_page.gettbody()
                cells = []
                index = 0
        cat_page.getrow(cells)
        cells = []
        cat_page.closetab()
#       cat_page.getlinkbar()
        cat_page.makelinkbar(self.nobjects, self.overview)
        cat_page.writetail()
        cat_page.closefile()

        return self.page_name


