"""
$Revision: 1.6 $ $Date: 2010-08-25 06:29:15 $
Author: Martin Kuemmel
Affiliation: Space Telescope - European Coordinating Facility
"""
__author__ = "Martin Kuemmel <mkuemmel@eso.org>"
__date__ = "$Date: 2010-08-25 06:29:15 $"
__version__ = "$Revision: 1.6 $"
__credits__ = """This software was developed by the ACS group of the Space Telescope -
European Coordinating Facility (ST-ECF). The ST-ECF is a department jointly
run by the European Space Agency and the European Southern Observatory.
It is located at the ESO headquarters at Garching near Munich. The ST-ECF
staff supports the European astronomical community in exploiting the research
opportunities provided by the earth-orbiting Hubble Space Telescope.
"""

import sys, os, os.path, string, math
import aXe2html.irafims.irafimtools
import aXe2html.sexcat.sextractcat
import aXe2html.spectra.plot_spc
import aXe2html.utils.namings, aXe2html.utils.namings
import aXe2html.spectra.outdata
import aXe2html.utils.fitsslicer
import pyfits

class ImPlotMaker:
    """
    The workhorse object which makes all images, that means the direct
    stamp image, the grisma stamp image as well as the the counts plot
    and the flux plot. Does that by creating the according instances
    of plot objects and executing the relevant methods.
    """
    def __init__(self, settings):
        """
        Input:
            settings - dictionary with all settings

        Return:
            -

        Description:
            Transforms all necessary settings in the input
            into instance variables. Those can be accessed
            by each instance and used to create other
            objects or calling other methods.
        """
        self.direct_image = settings["DIRIMAGE"]
        self.grism_image  = settings["STPIMAGE"]
        self.spec_table   = settings["SPCTABLE"]
        self.sextract_cat = settings["SEXTRCAT"]
        self.start_object = settings["STARTOBJ"]
        self.nobjects     = settings["NUMBEOBJ"]
        self.specwidth    = settings["SPECWIDT"]
        self.rowsize      = settings["DIMWIDTH"]
        self.sortcolumn   = settings["SORTCOLU"]
        self.errorbar     = settings["ERRORBAR"]
        self.sortorder    = settings["ASCENDIN"]
        self.outpath      = settings['OUTPPATH']
        self.imgpath      = settings['IMAGPATH']
        self.fitspath     = settings['FITSPATH']
        self.dataoutp     = settings['DATAOUTP']
        self.datapath     = settings['DATAPATH']
        self.marksource   = settings['MARKSOUR']
        self.wavlower     = settings['WAVLOWER']
        self.wavupper     = settings['WAVUPPER']
        self.mag_wave     = settings['MAGLAMBD']

    def _get_specmap_entry(self, axedata):
        """
        """
        # define the extension name
        beamID = 'BEAM_%iA' % axedata['NUMBER']

        # define the spectrum name
        speName = 'spec%s.fits' % beamID

        # define the path to the spectrum
        specPath = os.path.join(self.outpath, self.fitspath, speName)

        # return the beamID and the path
        return beamID, specPath

    def _get_grismmap_entry(self, axedata):
        """
        """
        # define the extension name
        beamID = 'BEAM_%iA' % axedata['NUMBER']

        # define the spectrum name
        griName = 'grism%s.fits' % beamID

        # define the path to the spectrum
        griPath = os.path.join(self.outpath, self.fitspath, griName)

        # return the beamID and the path
        return beamID, griPath

    def makeimplots(self):
        """
        Input:
           -

        Return:
           -

        Description:
            Does the necessary steps to create the different images.
            Calls the sextractor catalog and makes the necessary
            steps (perhaps ordering and so on) to get the requested
            sequence of objects to be worked on.
            For each object the aXe data is extracted, and an
            according object is created for each image (2xstamp
            + 2xplot).
            The actual images are created by executing the methods
            of that objects.
        """
        import shutil

        # make an empty dictionary
        spec_map = {}
        grism_map = {}

#--------------------------------------------------------------------
#
#   Open the Sextracor catalog and the direct image.
#   Sort the catalog if requested
        dir_ima = aXe2html.irafims.irafimtools.IrafImage(self.direct_image)
        sex_cat = aXe2html.sexcat.sextractcat.SexCat(self.sextract_cat)
        success = sex_cat.setaxecols(self.mag_wave)
        if success > 0:
            return 100
        if self.sortcolumn != None:
            if self.sortorder == 'YES':
                success = sex_cat.setorder(self.sortcolumn, 0)
            else:
                success = sex_cat.setorder(self.sortcolumn, 1)
            if success > 0:
                return 101
#
#--------------------------------------------------------------------

#--------------------------------------------------------------------
#
#   Determine the number of objects and derive
#   the number of html-pages which are to be created
#
        allobjects = sex_cat.getnrows()
        if self.nobjects == -1:
           self. nobjects   = allobjects
        elif self.nobjects > allobjects:
            print 'Sorry, I can give you only ', allobjects, ' objects.'
            print 'The catalog is not large enough!'
            self.nobjects = allobjects

#
#--------------------------------------------------------------------

        # check for quantitative contamination
        quant_cont = has_quant_contmodel(self.spec_table)

        # go over all objects        
        for oneobject in range(self.nobjects):

            #   Get the sextractor information on
            #   the current object
            actobject = oneobject + self.start_object - 1
            if self.sortcolumn != None:
                axedata = sex_cat.getaxesortdata(actobject)
            else:
                axedata = sex_cat.getaxedata(actobject)

            # get the name key and the name value
            spec_key, spec_value = self._get_specmap_entry(axedata)

            # make an entry in the name map
            spec_map[spec_key] = spec_value

            # make an entry in the name map
            spec_map[spec_key] = spec_value

            # get the name key and the name value
            grism_key, grism_value = self._get_grismmap_entry(axedata)

            # make an entry in the name map
            grism_map[grism_key] = grism_value

        # slice the SPC file to individual spectra
        spec_slicer = aXe2html.utils.fitsslicer.FitsSlicer(self.spec_table, name_map=spec_map, verbose=False)
        spec_slicer.run()
        del spec_slicer

        # slice the STP file to individual spectra
        stp_slicer = aXe2html.utils.fitsslicer.FitsSlicer(self.grism_image, name_map=grism_map, verbose=False)
        stp_slicer.run()
        del stp_slicer


#--------------------------------------------------------------------
#
#   Make a big loop over all objects.
#   Do everything within that big loop
#
        for oneobject in range(self.nobjects):

#--------------------------------------------------------------------
#
#   Get the sextractor information on
#   the current object
            actobject = oneobject + self.start_object - 1
            if self.sortcolumn != None:
                axedata = sex_cat.getaxesortdata(actobject)
            else:
                axedata = sex_cat.getaxedata(actobject)
#--------------------------------------------------------------------

            # define the beam ID
            beamID = 'BEAM_%iA' % axedata['NUMBER']

#--------------------------------------------------------------------
#
#   Derive the proper names for the images and grafics,
#   then make them

            # prepare the direct image stamp
            area_string = dir_ima.set_areastring(axedata['X_IMAGE'],axedata['Y_IMAGE'],self.rowsize)
            
            # check whether something can be done
            if area_string != None:    

                # make the direct image stamp
                imname = dir_ima.extractstamp(axedata['X_IMAGE'],axedata['Y_IMAGE'],self.rowsize, self.marksource, \
                                              aXe2html.utils.namings.DirectStampName(axedata['NUMBER'], \
                                                                                     self.marksource))

                # move the stamp to the right location
                fullname = os.path.join(self.outpath, self.imgpath, imname)
                os.rename(imname, fullname)

                # comment the next lines if the
                # fits stamps shall NOT be saved
                #
                name_object = aXe2html.utils.namings.DirectStampName(axedata['NUMBER'], self.marksource)
                fitsname = name_object.getfitsname()
                fullname = os.path.join(self.outpath, self.fitspath, fitsname)
                os.rename(fitsname , fullname)

            else:
                
                # generate and move the default image
                fromstring = aXe2html.utils.namings.DefaultImages().getdefdimname()
                tostring   = os.path.join(self.outpath, self.imgpath, aXe2html.utils.namings.DirectStampName(axedata['NUMBER'], self.marksource).getdefimname())
                shutil.copyfile(fromstring, tostring)

            fitsstamp = '%s[%s]' % (grism_map[beamID], beamID)
            gri_ima = aXe2html.irafims.irafimtools.IrafImage(fitsstamp)
            imname = gri_ima.extspecstamp( self.specwidth, self.rowsize, \
                                           aXe2html.utils.namings.GrismStampName(axedata['NUMBER']))
            if imname != 'noimage':
                fullname = os.path.join(self.outpath, self.imgpath, imname)
                os.rename(imname, fullname)
            else:
                fromstring = aXe2html.utils.namings.DefaultImages().getdefgriname()
                tostring   = os.path.join(self.outpath, self.imgpath, aXe2html.utils.namings.GrismStampName(axedata['NUMBER']).getdefimname())
                shutil.copyfile(fromstring, tostring)

            # check whether the spectrum exists
            if os.path.isfile(spec_map[beamID]):

                # make a plot object
                plotobj = aXe2html.spectra.plot_spc.MatPlotter(spec_map[beamID], beamID, \
                            aXe2html.utils.namings.CountsPlotName(axedata['NUMBER'],self.errorbar),\
                            aXe2html.utils.namings.FluxPlotName(axedata['NUMBER'],self.errorbar),\
                            self.errorbar, self.wavlower, self.wavupper, quant_cont)

                try:
                    # try to do the plotting
                    imnames = plotobj.plotAll()

                    # move the plots to their location
                    for name in imnames:
                        fullname = os.path.join(self.outpath, self.imgpath, name)
                        os.rename(name, fullname)
            
                except:
                    # copy the empty plots
                    fromstring = aXe2html.utils.namings.DefaultImages().getdefploname()
                    tostring   = os.path.join(self.outpath, self.imgpath, aXe2html.utils.namings.CountsPlotName(axedata['NUMBER'],self.errorbar).getdefimname())
                    shutil.copyfile(fromstring, tostring)

                    fromstring = aXe2html.utils.namings.DefaultImages().getdefploname()
                    tostring   = os.path.join(self.outpath, self.imgpath, aXe2html.utils.namings.FluxPlotName(axedata['NUMBER'],self.errorbar).getdefimname())
                    shutil.copyfile(fromstring, tostring)
            else:
                # copy the empty plots
                fromstring = aXe2html.utils.namings.DefaultImages().getdefploname()
                tostring   = os.path.join(self.outpath, self.imgpath, aXe2html.utils.namings.CountsPlotName(axedata['NUMBER'],self.errorbar).getdefimname())
                shutil.copyfile(fromstring, tostring)

                fromstring = aXe2html.utils.namings.DefaultImages().getdefploname()
                tostring   = os.path.join(self.outpath, self.imgpath, aXe2html.utils.namings.FluxPlotName(axedata['NUMBER'],self.errorbar).getdefimname())
                shutil.copyfile(fromstring, tostring)

            # check whether a data file
            # is requested
            if self.dataoutp == 'YES':
                
                # check whether the file exists
                if os.path.isfile(spec_map[beamID]):
                    # create output spectrum object
                    dataobj = aXe2html.spectra.outdata.OutSpect(spec_map[beamID], axedata['NUMBER'],\
                                            aXe2html.utils.namings.SpectDataName(axedata['NUMBER'], 'TXT'))

                    # write spectrum to fits
                    datafile = dataobj.fits2string()

                    # check for success
                    if datafile:

                        # determine the full path to the data file
                        fullname = os.path.join(self.outpath, self.datapath, datafile)
                        # move the data file to the right location
                        os.rename(datafile, fullname)

                    else:
                        print ' Object: ',  str(axedata['NUMBER']), ' is not in the SPC file!'

                    # mop up
                    del dataobj

                else:
                    print ' Object: ',  str(axedata['NUMBER']), ' is not in the SPC file!'

            # delete the sliced version
            # of the spectrum and the grism stamp
            if os.path.isfile(spec_map[beamID]):
                os.unlink(spec_map[beamID])
            if os.path.isfile(grism_map[beamID]):
                os.unlink(grism_map[beamID])

            print '>>> Finished object ', str(axedata['NUMBER']), '>>>'
#--------------------------------------------------------------------

        return 0

def has_quant_contmodel(spc_name):
    """
    Input:
    cont_model - the name of the emission model

    Return:
    isquant - 1 in case of a quantitative model,
    0 if not

    Description:
    The module checks whether a contamination model
    is quantitative or not.
    """
    quant_models = ['gauss', 'fluxcube']

    isquant=0

    ifile = pyfits.open(spc_name)
    ihead = ifile[0].header

    if ihead.has_key('CONTAM'):

        cont_model = ihead['CONTAM']

        for i in range(len(quant_models)):
            if string.lower(string.strip(cont_model)) == quant_models[i]:
                isquant=1
                break

    return isquant
