"""
$Revision: 1.4 $ $Date: 2010-08-25 06:29:15 $
Author: Martin Kuemmel
Affiliation: Space Telescope - European Coordinating Facility
"""
__author__ = "Martin Kuemmel <mkuemmel@eso.org>"
__date__ = "$Date: 2010-08-25 06:29:15 $"
__version__ = "$Revision: 1.4 $"
__credits__ = """This software was developed by the ACS group of the Space Telescope -
European Coordinating Facility (ST-ECF). The ST-ECF is a department jointly
run by the European Space Agency and the European Southern Observatory.
It is located at the ESO headquarters at Garching near Munich. The ST-ECF
staff supports the European astronomical community in exploiting the research
opportunities provided by the earth-orbiting Hubble Space Telescope.
"""

from pyraf import iraf
import pyfits
import math, string, os
from aXe2html.utils.mathlib import computestat, rescale_int, \
     getscales, getpercentiles, filtZEROminusINF
from aXe2html.utils.fitsheader import getallkey
import previewplot

class IrafImage:
    def __init__(self, filename):
        self.imname = filename

    def set_areastring(self, xcen, ycen, size):
        """
        Input:
            xcen    - the xcoordinate of the object 
            ycen    - the ycoordinate of the object
            size    - the size (in pixels) of the stamp image

        Return:
            self.areastring - the IRAF area string for the stamp

        Description:
            This method defines the IRAF area string that can
            be used in 'imcopy' to extract the stamp image.
        """
        headerdict = getallkey(self.imname)


        # make sure something is to be done
        if xcen > 0 and ycen > 0 and xcen < int(headerdict["NAXIS1"]) and ycen < int(headerdict["NAXIS2"]):
            
            xstart = int(math.floor(xcen-(size-1)/2.0+0.5))
            xend   = int(math.floor(xcen+(size-1)/2.0+0.5))
            ystart = int(math.floor(ycen-(size-1)/2.0+0.5))
            yend   = int(math.floor(ycen+(size-1)/2.0+0.5))
    
            self.xstamp = int(math.floor(size/2.0+0.5))
            self.ystamp = int(math.floor(size/2.0+0.5))
    
            #  find the  right borders of the stamp image 
            if xstart < 1:
                xstart = 1
                xend = size
                self.xstamp = int(math.floor(xcen))
                markobj = 'yes'
            if ystart < 1:
                ystart = 1
                yend = size
                self.ystamp = int(math.floor(ycen))
                markobj = 'yes'
            if xend > int(headerdict["NAXIS1"]):
                xend = int(headerdict["NAXIS1"])
                xstart = xend-size+1
                self.xstamp = int(math.floor(size-(int(headerdict["NAXIS1"])-xcen)+0.5))
                markobj = 'yes'
            if yend > int(headerdict["NAXIS2"]):
                yend = int(headerdict["NAXIS2"])
                ystart = yend-size+1
                self.ystamp = int(math.floor(size-(int(headerdict["NAXIS2"])-ycen)+0.5))
                markobj = 'yes'
                
            # compose the area string
            self.areastring = '[%i:%i,%i:%i]' % (xstart, xend, ystart, yend)

        # nothing can be done
        else:
            
            # set everything to None
            self.xstamp     = None
            self.ystamp     = None
            self.areastring = None


        # return the area string
        return self.areastring

    def extractstamp(self, xcen, ycen, size, mark, outname):
        """
        Input:
            xcen    - the xcoordinate of the object 
            ycen    - the ycoordinate of the object
            size    - the size (in pixels0 of the stamp image
            outname - naming object to get the names straight

        Return:
            outname.getdefimname() - then name of the stamp image

        Description:
            This method extracts the stamp image of size 'size' around
            'xcen', 'ycen' from the direct image.
            If the object can not be centered, the stamp starts at the
            first row/column or ends at last row/column.
            Then that subimage is extraced and turned such that the
            extracted data have the right orientation as a PIL image.
            The data is extraced and the data values sclaed to give
            an appropriate contrast in the image.
            From that scaled data values the PIL-image is created,
            the red cross over the main object is drawn and then
            the picture is stored in the right format.
        """
        if os.path.isfile(outname.getfitsname()):
                os.unlink(outname.getfitsname())
        if os.path.isfile(outname.getdefimname()):
                os.unlink(outname.getdefimname())

        # copy the stamp to a temporary file and trnasform it such that the data
        # have the right orientation for the PIL-image
        inname = self.imname + self.areastring
        iraf.imcopy(inname, outname.getfitsname(), Stdout=1)

        # make the web-image
        pplot = previewplot.PowerShotMp(outname.getfitsname(), 0)
        if mark == 'YES' and self.xstamp != None and self.ystamp != None:
            pplot.make_cross(self.xstamp,self.ystamp)
        pplot.plot(outname.getdefimname())

        # switch this in if the fits stamps shall
        # NOT be saved
        #        if os.path.isfile(outname.getfitsname()):
        #                os.unlink(outname.getfitsname())

        # store the object position in the fits
        if self.xstamp != None and self.ystamp != None:
            img = pyfits.open(outname.getfitsname(), 'update')
            imh = img[0].header
            imh.update('xpos', self.xstamp, 'x-position of source on image')
            imh.update('ypos', self.ystamp, 'y-position of source on image')
            img.flush()
            img.close()

        # retrun the web-image name
        return outname.getdefimname()


    def extspecstamp(self, width, heigth, outname):
        """
        Input:
            width    - the width of the stamp image
            heigth   - the height of the stamp image
            outname  - naming object to get the names straight

        Return:
            outname.getdefimname() - then name of the stamp image

        Description:
            The method takes the whole aXe grism stamp image.
            The width of the image in the html-document is fixed,
            the height is up to now only a dummy.
            If the width of the fits image is in a certain corridor
            (152 < width < 172), nothing is done. Otherwise the
            image is scaled that its width matches the specified
            width.
            The extracted subimage is turned such that the
            extracted data have the right orientation as a PIL image.
            The data is extracted and the data values scaled to give
            an appropriate contrast in the image.
            From that scaled data values the PIL-image is created,
            the red cross over the main object is drawn and then
            the picture is stored in the right format.
        """
        # resolve the header keywords
        headerdict = getallkey(self.imname)

        # save the fits name
        inname = self.imname

        # guess the next lines are not necessary
        # anymore
        point = string.find(inname,'.fits')
        rootname = inname[0:point+5]
        extname = inname[point+6:len(inname)-1]

        if not headerdict.has_key("NAXIS1"):
            return 'noimage'

        if int(headerdict["NAXIS1"]) < 152 or int(headerdict["NAXIS1"]) > 172:
            scalefac = width/float(headerdict["NAXIS1"])
        else:
            scalefac = 1.0

        if os.path.isfile(outname.getfitsname()):
                os.unlink(outname.getfitsname())
        if os.path.isfile(outname.getdefimname()):
                os.unlink(outname.getdefimname())

        if scalefac != 1.0:
            iraf.magnify(inname, outname.getfitsname(), xmag=scalefac, \
                         ymag=scalefac, Stdout=1)
        else:
            iraf.imcopy(inname, outname.getfitsname(), Stdout=1)

        # genrate the web-image
        pplot = previewplot.PowerShotMp(outname.getfitsname(), 0)
        pplot.plot(outname.getdefimname())

        # delete the original fits
        if os.path.isfile(outname.getfitsname()):
                os.unlink(outname.getfitsname())
    
        # return the web-image name
        return outname.getdefimname()
