"""
$Revision: 1.6 $ $Date: 2011-06-03 13:44:15 $
Author: Martin Kuemmel
Affiliation: Space Telescope - European Coordinating Facility
"""
__author__ = "Martin Kuemmel <mkuemmel@eso.org>"
__date__ = "$Date: 2011-06-03 13:44:15 $"
__version__ = "$Revision: 1.6 $"
__credits__ = """This software was developed by the ACS group of the Space Telescope -
European Coordinating Facility (ST-ECF). The ST-ECF is a department jointly
run by the European Space Agency and the European Southern Observatory.
It is located at the ESO headquarters at Garching near Munich. The ST-ECF
staff supports the European astronomical community in exploiting the research
opportunities provided by the earth-orbiting Hubble Space Telescope.
"""

"""
03-June-2011 H.Bushouse STScI v1.6
Modified the prepare_data method to exclude bad pixels flagged with values
of -1e6, which, if not ignored, can cause the image scaling on the html
pages to be incorrect. Trac ticket #714.
"""

try:
    reload(matplotlib)
except NameError:
    import matplotlib

matplotlib.use('Agg')
from matplotlib import pylab


class PowerShotMp(object):
    """
    General class for plotting images
    """
    def __init__(self, infile, hdunum=None, qmin=1, qmax=99, invert=True):
        """
        Initializes the class
        """
        import pylab
        # store the file name
        self.infile = infile

        # save the parameter hdunum
        if hdunum == None:
            self.hdunum = 0
        else:
            self.hdunum = hdunum

        # save the parameter qmin/qmax
        self.qmin = qmin
        self.qmax = qmax

        # save other input
        self.invert = invert

        # load, flatten and sort the image data
        flat_data, naxis1, naxis2 = self._prepare_data(infile, hdunum)

        # generate the figure
        self.figsize=int(float(naxis1)), int(float(naxis2))
        self.preview_figure = pylab.figure(num=1, figsize=self.figsize)
        pylab.clf()

        # save the shape
        self.shape = (naxis1, naxis2)
        
        # check whether there is data
        if len(flat_data) > 0:
            # get the minimum and maximum values
            self.minval = self._get_qval(flat_data, self.qmin)
            self.maxval = self._get_qval(flat_data, self.qmax)
            del flat_data
            flat_data = None
        else:
            # give default values
            self.minval = 0.0
            self.maxval = 0.0
            del flat_data
            flat_data = None

        # load, flatten and sort the image data
        self.data = self._get_data(infile, hdunum)

        # generate the axex
        self.plot_axes  = self.preview_figure.add_axes([0.,0.,1.0,1.0])
        self.addto_figure(self.plot_axes)

    def _get_qval(self, in_data, qvalue):
        """
        Determine the data value for a quantile
        """
        # store the data length
        ndata = len(in_data)

        # get the right index
        ind_qval = int(float(ndata)*qvalue/100.0)

        # secure the index
        if ind_qval < 0 :
            ind_qval=0
        if ind_qval > ndata-1:
            ind_qval=ndata-1

        # get the desired value
        data_value = in_data[ind_qval]

        # save memory
        del in_data
        in_data = None

        # return the data point
        return data_value

    def _prepare_data(self, infile, hdunum):
        """
        Load, flatten and sort image data
        """
        import pyfits
        import numpy

        # initialize a dummy array
        an_arr = []

        # load the image
        img = pyfits.open(infile)

        # load the data of the specified extension
        img_data = img[hdunum].data

        # derive the data dimension
        img_shape = img_data.shape

        # derive the number of pixels
        npix = img_shape[0] * img_shape[1]

        # flatten the data array
        flat_data = numpy.reshape(img_data, (npix))

        # select only non-zero values
        flat_data = numpy.array(flat_data[numpy.where(flat_data != 0.0)])

        # select only non-flagged values; ticket #714 (3-June-2011)
        goodvals = flat_data >= -1e5
        if goodvals.any() != False:
           flat_data = flat_data[goodvals]

        # sort the data
        sort_data = numpy.sort(flat_data)

        # delete the array
        del img_data
        del flat_data
        img_data = None
        flat_data = None

        # close the image
        img.close()
        img = None

        # return the sorted array
        return sort_data, img_shape[1], img_shape[0]

    def _get_data(self, infile, hdunum):
        """
        Load, flatten and sort image data
        """
        import pyfits

        # initialize a dummy array
        an_arr = []

        # load the image
        img = pyfits.open(infile)

        # load the data of the specified extension
        img_data = img[hdunum].data.copy()

        # close the fits file
        img.close()

        # return the sorted array
        return img_data

    def _mopup(self):
        """
        Delete some memory
        """
        # delete the data
        del self.data
        self.data = None

    def addto_figure(self, plot_axes):
        """
        Add the image plot to the figure
        """
        import pylab

        # do the basic plot
        if self.invert:
            # either inverted
            plot_axes.imshow(-1.0 * self.data, origin='lower', aspect=None, interpolation='nearest',
                             extent=[0,self.shape[0], 0,self.shape[1]], cmap=matplotlib.cm.gray, vmin=-1.0 * self.maxval, vmax=-1.0 * self.minval)
        else:
            # or normal
            plot_axes.imshow(self.data, origin='lower', aspect=None, interpolation='nearest',
                             extent=[0,self.shape[0], 0,self.shape[1]],cmap=matplotlib.cm.gray, vmin=self.minval, vmax=self.maxval)

    def make_cross(self, xpos, ypos):
        """
        Plot the finding cross
        """
        # define the size of the cross
        size = 10

        # plot the starting and end
        # positions of the cross
        xstart = xpos - size
        xend   = xpos + size
        ystart = ypos - size
        yend   = ypos + size

        # plot the trace lines
        self.plot_axes.plot([xstart, xend], [ypos, ypos], linestyle='-', color='r', linewidth=100.0)
        self.plot_axes.plot([xpos, xpos], [ystart, yend], linestyle='-', color='r', linewidth=100.0)

        # set the right range
        self.plot_axes.axis([0,self.shape[0], 0, self.shape[1]])

    def plot(self, plot_name):
        """
        Finish and save the plot
        """
        # set all the tickmarks to NULL
        self.plot_axes.xaxis.set_minor_locator(matplotlib.ticker.NullLocator())
        self.plot_axes.yaxis.set_minor_locator(matplotlib.ticker.NullLocator())
        self.plot_axes.xaxis.set_major_locator(matplotlib.ticker.NullLocator())
        self.plot_axes.yaxis.set_major_locator(matplotlib.ticker.NullLocator())

        # save the figure
        self.preview_figure.savefig(plot_name, dpi=1)

        # close all pylab figures
        pylab.close('all')
