"""
$Revision: 1.4 $ $Date: 2010-08-25 06:47:17 $
Author: Martin Kuemmel
Affiliation: Space Telescope - European Coordinating Facility
"""
__author__ = "Martin Kuemmel <mkuemmel@eso.org>"
__date__ = "$Date: 2010-08-25 06:47:17 $"
__version__ = "$Revision: 1.4 $"
___credits__ = """This software was developed by the ACS group of the Space Telescope -
European Coordinating Facility (ST-ECF). The ST-ECF is a department jointly
run by the European Space Agency and the European Southern Observatory.
It is located at the ESO headquarters at Garching near Munich. The ST-ECF
staff supports the European astronomical community in exploiting the research
opportunities provided by the earth-orbiting Hubble Space Telescope.
"""

import os, string, math
from   pyraf   import iraf
import pyfits
import aXe2html.asciidata

class OutSpect:
    """
    Class for the data file output
    """
    def __init__(self, inname, beamID, namobj):
        """
        Input:
            inname - the full spectrum table name 
            namobj - name object for the data file
            
        Return:
            -

        Description:
            Creates a class instances.
        """
        # parameters --> class variables
        self.inname = inname
        self.beamID = beamID
        self.namobj = namobj

        # fixed temporary name
        #self.tmpname = 'tmptab.fits'

        # name and description for columns to be
        # written to the data file
        self.fields = [('LAMBDA', 'Column1: LAMBDA'),
                       ('COUNT', 'Column2: COUNT'),
                       ('ERROR','Column3: ERROR'),
                       ('FLUX','Column4: FLUX'),
                       ('FERROR','Column5: FERROR'),
                       ('CONTAM','Column6: CONTAM')
                       ]

    def write_to_gzip(self, adata, fname):
        import gzip

        gfile = gzip.open(fname, 'w')
        gfile.write(str(adata)+'\n')
        gfile.close()
    
    def write_to_bz2(self, adata, fname):
        import bz2

        bz2file = bz2.BZ2File(fname, 'w')
        bz2file.write(str(adata)+'\n')
        bz2file.close()  

    def write_to_txt(self, adata, fname):

        tfile = open(fname, 'w')
        tfile.write(str(adata)+'\n')
        tfile.close()            

        
    def transfer_data(self, fdata, adata):
        """
        Input:
            fdata - the fits object
            adata - the asciidata object
            
        Return:
            -
            
        Description:
            Transfers the data from the fits file to the asciidata
            object. Fills also the header with the field names
            and comments defined in the object instance.
        """

        # go over all rows
        a_index = 0
        for index in range(adata.nrows):

            # check for useful data
            if fdata[1].data.field('COUNT')[index] != 0.0 \
               and fdata[1].data.field('ERROR')[index] != 0.0:
            
                # go over all fields defined
                # in the class
                findex = 0
                for field in self.fields:

                    # copy an entry fro the fits-file to the
                    # asciidata table
                    adata[findex][a_index] = fdata[1].data.field(field[0])[index]
                    findex+=1

                # enhance the index for the asciidata table
                a_index+=1
                
        # delete not used rows
        adata.delete(a_index, adata.nrows)

        # make a commment in the header
        adata.header.append('')

        # fill the header
        for field in self.fields:
            adata.header.append(field[1])

        if self.inname.find('[') > -1:
            fname = string.split(self.inname, '[')[0]
            adata.header.append('')
            adata.header.append('SPC-file: '+fname)
        adata.header.append('BeamID  : '+str(self.beamID))

        # make a commment in the header
        adata.header.append('')

    def fits2string(self):
        """
        Input:
            -
            
        Return:
            - the name of the zipped data file

        Description:
            Converts the spectrum in the fits-table to the data file.
            Controls all actions from the creation of a tmp-file
            over the opening of the asciidata class to the writing
            into a zipped file.
        """
        import gzip

        # check whether the table is there
        if not os.path.isfile(self.inname):
            # return None if not
            return None

        # open the fits table
        fdata = pyfits.open(self.inname)

        # create the asciidata object
        adata = aXe2html.asciidata.create(len(self.fields), fdata[1].data.shape[0])

        # transfer the data from the fits-file to the
        # asciidata object
        self.transfer_data(fdata, adata)

        # write the asciidata object to the file
        #self.write_to_gzip(adata, self.namobj.getdefdataname())
        #    write_to_bz2(adata, SpectDataName(20, 'BZ2').getdefimname())
        self.write_to_txt(adata, self.namobj.getdefdataname())

        # close the fits-file 
        fdata.close()

        # return the name of the data file
        return self.namobj.getdefdataname()
