"""
$Revision: 1.5 $ $Date: 2010-08-25 06:47:17 $
Author: Martin Kuemmel
Affiliation: Space Telescope - European Coordinating Facility
"""
__author__ = "Martin Kuemmel <mkuemmel@eso.org>"
__date__ = "$Date: 2010-08-25 06:47:17 $"
__version__ = "$Revision: 1.5 $"
___credits__ = """This software was developed by the ACS group of the Space Telescope -
European Coordinating Facility (ST-ECF). The ST-ECF is a department jointly
run by the European Space Agency and the European Southern Observatory.
It is located at the ESO headquarters at Garching near Munich. The ST-ECF
staff supports the European astronomical community in exploiting the research
opportunities provided by the earth-orbiting Hubble Space Telescope.
"""

import sys, os, string
import pyfits, math

try:
    reload(matplotlib)
except NameError:
    import matplotlib

from matplotlib import pylab


import aXe2html.utils.mathlib

FLUX_SCALE  = 1.0e+18

class MatPlotter:

    def __init__(self, fits_nam, fits_ext, cnamobj=None, fnamobj=None,
                 errorb='YES', wav_min=5500.0, wav_max=10500.0, quant_cont=1):
        """
        Input:
            fits_nam   - the name of the SPC table
            fits_ext   - the extension (in letters)
            cnamobj    - name instance for count plot
            fnamobj    - name instance for flux plot
            errorb     - boolean to reuqest error bar
            wav_min    - the minimum wavelength to plot
            wav_max    - the maximum wavelength to plot
            quant_cont - boolean for quantitative contamination

        Return:
            plot_names - list wiht the name sof the plots

        Description:
            Initializes the class, stores the variables
        """

        self.fits_nam = fits_nam
        self.fits_ext  = fits_ext
        self.cnamobj   = cnamobj
        self.fnamobj   = fnamobj
        self.errorb    = errorb
        self.wav_min   = wav_min
        self.wav_max   = wav_max
        self.qcont     = quant_cont

    def plotAll(self):
        """
        Input:
            -

        Return:
            plot_names - list wiht the name sof the plots

        Description:
            Controls the whole process to create the plots
            for one extension of a SPC table.
            It extracts the data, cleans the data and
            creates both, the counts plot and the flux plot.
        """
        # initialize the return structure
        plot_names = []

        # extract the plot data
        pdata = self.get_plotdata(self.fits_nam, self.fits_ext,
                                  self.wav_min, self.wav_max)

        # remove leading and
        # trailing zeros
        self.clean_plotarrays(pdata)

        # check whether there is still
        # data to be plotted
        if len(pdata['LAMBDA']) < 1:
            # raise an exception
            # if the plot array is empty
            raise Exception('No data left!')

        # make the cont plot
        plot_names.append(self.make_cplot(pdata, self.wav_min, self.wav_max, self.errorb))

        # make the flux plot
        plot_names.append(self.make_fplot(pdata, self.wav_min, self.wav_max, self.errorb, self.qcont))

        # return the list with the plot names
        return plot_names


    def get_plotdata(self, fits_nam, fits_ext,
                     lambda_min=0.0, lambda_max=1.0e+10):
        """
        Input:
            fits_nam   - the name of the SPC table
            fits_ext   - the extension (in letters)
            lambda_min - the minimum wavelength to plot
            lambda_max - the maximum wavelength to plot

        Return:
            pdata - the data extracted from the SPC table

        Description:
            Extracts the data to be plotted from the correct
            extension of the SPC file. Stores and returns
            the data in a cusmtom, easy data structure.
        """

        # initialize the arrays
        # to contain the data
        pdata = {}
        lambd = []
        count = []
        error = []
        flux  = []
        ferro = []
        conta = []

        # open the fits, and extract
        # the data from the right extention
        ffile = pyfits.open(fits_nam)
        #fdata = ffile[fits_ext].data
        fdata = ffile[1].data

        # go over all rows
        for index in range(len(fdata)):
            # check whether the wavelength fits in
            # the correct range
            if fdata[index].field('LAMBDA') > lambda_min and \
                   fdata[index].field('LAMBDA') < lambda_max:

                # copy the data to the arrays
                lambd.append(fdata[index].field('LAMBDA'))
                count.append(fdata[index].field('COUNT'))
                error.append(fdata[index].field('ERROR'))
                flux.append(fdata[index].field('FLUX'))
                ferro.append(fdata[index].field('FERROR'))
                conta.append(fdata[index].field('CONTAM'))

        # put the arrays as dictionary entries
        # in a dixtionary
        pdata['LAMBDA'] = lambd
        pdata['COUNT']  = count
        pdata['ERROR']  = error
        pdata['FLUX']   = flux
        pdata['FERROR'] = ferro
        pdata['CONTAM'] = conta

        # return the dictionary
        return pdata

    def make_cplot(self, pdata, wav_min, wav_max, errorb):
        """
        Input:
            pdata   - the data to be plotted
            wav_min - lower wavelength plotted
            wav_max - upper wavelength plotted
            errorb  - boolean to request error bars

        Return:
            defname - the name of the plot created

        Description:
            Creates a count plot with error bars if
            requested and transforms the plot into the
            desired format.
        """
        from matplotlib.ticker import MultipleLocator

        # get the plot names for png format
        # and for the desired format
        pngname = self.cnamobj.getpngname()
        defname = self.cnamobj.getdefimname()

        # determine the dimensions in y-direction
        ydims = self.setydimensions(pdata['COUNT'], pdata['ERROR'])

        # set the correct figure size (320x240pt)
        matplotlib.pylab.figure(num=1, figsize=(3.2,2.4), dpi=100)
        pylab.clf()

        # position the axis in the plot
        matplotlib.pylab.axes([.2,.17,.78,.8])

        # label the x-axis
        pylab.figtext(.6,.05,r'$wavelength\ [\AA]$', ha='center', va='center', size=12)

        # label the y-axis
        pylab.figtext(.05,.5,r'$counts [e^{-}/s]$', ha='center', va='center',rotation=90, size=12)

        # check for error bars
        if errorb == 'YES':
            # plot data and errors
            pylab.errorbar(pdata['LAMBDA'],pdata['COUNT'],pdata['ERROR'],None, 'ro', aa=False, mew=0, mec='r',
                           markersize=2, elinewidth=1)
        else:
            # plot the data only
            pylab.plot(pdata['LAMBDA'],pdata['COUNT'], 'ro',
                       aa=False, mew=0, mec='r', markersize=2)

        # set the tickmarks if necessary
        if wav_max-wav_min > 10000.0:
            maj_tick_size = None
            min_tick_size = None
            
        elif wav_max-wav_min > 5500.0:
            maj_tick_size = 2000
            min_tick_size = 1000

        elif wav_max-wav_min > 3000.0:
            maj_tick_size = 1000
            min_tick_size = 200
        else:
            maj_tick_size = None
            min_tick_size = None
        
        if maj_tick_size  and min_tick_size:
            # define and set the major tickmarks
            xmajor_locator = MultipleLocator(maj_tick_size)
            pylab.setp(pylab.gca().get_xaxis().set_major_locator(xmajor_locator))

            # define and set the minor tickmarks
            xminor_locator = MultipleLocator(min_tick_size)
            pylab.setp(pylab.gca().get_xaxis().set_minor_locator(xminor_locator))

        # set the size of the tickmarks
        pylab.setp( pylab.gca().get_xticklabels(), fontsize=10)
        pylab.setp( pylab.gca().get_yticklabels(), fontsize=10)
   
        # set the plot limits
        matplotlib.pylab.axis([wav_min, wav_max, ydims[0], ydims[1]])

        # save/render the plot and close it
        pylab.savefig(pngname, dpi=100)
        pylab.close('all')

        # check whether the desired
        # image format is 'png'
        if defname != pngname:
            # if not, then transform to
            # the desired format
            im = Image.open(pngname)
            im.save(defname)

            # delete the original image
            if os.path.isfile(pngname):
                os.unlink(pngname)

        # return the image name
        return defname

    def make_fplot(self, pdata, wav_min, wav_max, errorb, qcont):
        """
        Input:
            pdata   - the data to be plotted
            wav_min - lower wavelength plotted
            wav_max - upper wavelength plotted
            errorb  - boolean to request error bars
            qcont   - boolean to signal quant. contamination

        Return:
            defname - the name of the plot created

        Description:
            Creates a flux plot with contamination, error bars if
            requested and transforms the plot into the
            desired format.
        """
        from matplotlib.ticker import MultipleLocator

        # get the plot names for png format
        # and for the desired format
        pngname = self.fnamobj.getpngname()
        defname = self.fnamobj.getdefimname()

        # scale the y-values
        self.scaleyvalues(pdata['FLUX'], pdata['FERROR'],
                          pdata['CONTAM'], FLUX_SCALE)

        # determine the dimensions in y-direction
        ydims = self.setydimensions(pdata['FLUX'], pdata['FERROR'])

        # set the correct figure size (320x240pt)
        matplotlib.pylab.figure(num=2, figsize=(3.2,2.4), dpi=100)
        pylab.clf()

        # position the axis in the plot
        matplotlib.pylab.axes([.2,.17,.78,.8])

        # label the x-axis
        pylab.figtext(.6,.05,r'$wavelength\ [\AA]$', ha='center', va='center', size=12)

        # label the y-axis
        pylab.figtext(.05,.5,r'$flux/10^{-18} [erg/cm^{2}/s/\AA]$', ha='center', va='center',  rotation=90, size=12)

        # plot error bars if requested
        if errorb =='YES':
            pylab.errorbar(pdata['LAMBDA'],pdata['FLUX'],pdata['FERROR'],None, 'ro', aa=False,
                           mew=0, mec='r', markersize=2, elinewidth=1)
        else:
            pylab.plot(pdata['LAMBDA'],pdata['FLUX'], 'ro',
                       aa=False, mew=0, mec='r', markersize=2)


        # plot the contamination info
        if qcont:
            # for quantit. cont., just plot the contaminating flux
            pylab.plot(pdata['LAMBDA'],pdata['CONTAM'], 'bo',
                        aa=False, mew=0, markersize=2)
        else:
            # for geometrical contamination trim the data
            # to delete not contaminated points
            self.trim_contam(pdata)

            # if some points are still left, plot them
            # over in a different colour
            if len(pdata['LAMBDA']) > 0 and len(pdata['LAMBDA']) > 0:
                pylab.plot(pdata['LAMBDA'],pdata['FLUX'], 'bv',
                           aa=False, mew=0, markersize=2)

        # set the tickmarks if necessary
        if wav_max-wav_min > 10000.0:
            maj_tick_size = None
            min_tick_size = None

        if wav_max-wav_min > 5500.0:
            maj_tick_size = 2000
            min_tick_size = 500

        elif wav_max-wav_min > 3000.0:
            maj_tick_size = 1000
            min_tick_size = 200
        else:
            maj_tick_size = None
            min_tick_size = None
        
        # set the tickmarcks if necessary
        if maj_tick_size  and min_tick_size:
            # define and set the major tickmarks
            xmajor_locator = MultipleLocator(maj_tick_size)
            pylab.setp(pylab.gca().get_xaxis().set_major_locator(xmajor_locator))

            # define and set the major tickmarks
            xminor_locator = MultipleLocator(min_tick_size)
            pylab.setp(pylab.gca().get_xaxis().set_minor_locator(xminor_locator))

        # set the size of the tickmarks
        pylab.setp( pylab.gca().get_xticklabels(), fontsize=10)
        pylab.setp( pylab.gca().get_yticklabels(), fontsize=10)

        # set the axis boundaries
        matplotlib.pylab.axis([wav_min, wav_max, ydims[0], ydims[1]])

        # save/render the plot and close it
        pylab.savefig(pngname, dpi=100)
        pylab.close('all')

        # check whether the desired
        # image format is 'png'
        if defname != pngname:
            # if not, then transform to
            # the desired format
            im = Image.open(pngname)
            im.save(defname)

            # delete the original image
            if os.path.isfile(pngname):
                os.unlink(pngname)

        # return the image name
        return defname


    def trim_contam(self, pdata):
        """
        Input:
            pdata - the data extracted from a SPC table

        Return:
            -

        Description:
            Trims the whole data array down to contain only
            the entries which do have a reasonable entry for
            geometrical contamination. This prepares the
            data array to overplot only the contaminated
            flux values.
        """

        # set the indewx pointer to the last value
        lindex = len(pdata['CONTAM'])-1

        # go over all entries
        for index in range(len(pdata['CONTAM'])):

            # check for contamination entry
            if math.fabs(pdata['CONTAM'][lindex]) < .1:

                # delete not contaminated entries
                del pdata['CONTAM'][lindex]
                del pdata['FLUX'][lindex]
                del pdata['LAMBDA'][lindex]

            # set the counter
            lindex -= 1


    def setydimensions(self, yval, yerr):
        """
        Input:
            yval - the list with the y-data values
            err  - the list with the err values

        Return:
            -

        Description:
            This method determines the plotting range. The main
            driver for the plotting range is either the range
            of the data values of the median size of the errors.
            The bigger of the two is extended by 10% or 20%
            on both directions to get the minimum and maximum
            plotting values.
            This procedure has been empirically developped
            and produces fine results for aXe-data so far.
        """
        yreturn = []

        ystats = aXe2html.utils.mathlib.computestat(yval)
        yrange = ystats['max']-ystats['min']
        estats = aXe2html.utils.mathlib.getpercentiles(yerr,50,50)
        erange = 2.0*estats[0]

        if yrange > erange:
            ymin = ystats['min'] - 0.1*(ystats['max']-ystats['min'])
            ymax = ystats['max'] + 0.1*(ystats['max']-ystats['min'])
        else:
            yrange = erange
            ymin = ystats['ave'] - 1.2*estats[0]
            ymax = ystats['ave'] + 1.2*estats[0]

        self.ymin = ymin
        self.ymax = ymax
        self.yrange = yrange

        yreturn.append(ymin)
        yreturn.append(ymax)
        yreturn.append(yrange)

        return yreturn

    def scaleyvalues(self, yval, err, ycon, scale):
        """
        Input:
            yval - list with the y-data values
            err  - list with the error values
            ycon - list with the contaminated y-data values
            scale - the scaling factor to be applied to the data

        Return:
           -

        Description:
            Scales the whole y-data set by an arbitrary factor.
            Since the lists are only goven by reference, that
            is don 'on the fly' (bad style??).
        """
        for index in range(len(yval)):
            yval[index] = yval[index] * scale
            err[index]  = err[index] * scale
        for index in range(len(ycon)):
            ycon[index] = ycon[index] * scale

    def clean_plotarrays(self, pdata):
        """
        Input:
            pdata - the data extracted from a SPC-table

        Return:
            -

        Description:
            Cuts leading and traling zeros at both ends of
            the data arrays. The criterium is:
            counts == 0.0 and error == 0.0.
        """

        # cut trailing zeros
        dec = 1
        while dec == 1 and len(pdata['LAMBDA']) > 0:
            if abs(pdata['LAMBDA'][-1]) < 1.0e-30 or \
                   (abs(pdata['COUNT'][-1]) < 1.0e-30 and abs(pdata['ERROR'][-1]) < 1.0e-30):
                del pdata['LAMBDA'][-1]
                del pdata['COUNT'][-1]
                del pdata['ERROR'][-1]
                del pdata['FLUX'][-1]
                del pdata['FERROR'][-1]
                del pdata['CONTAM'][-1]
            else:
                dec=0

        # cut leading zeros
        dec = 1
        while dec == 1 and len(pdata['LAMBDA']) > 0:
            if abs(pdata['LAMBDA'][0]) < 1.0e-30 or \
                   (abs(pdata['COUNT'][0]) < 1.0e-30 and abs(pdata['ERROR'][0]) < 1.0e-30):
                del pdata['LAMBDA'][0]
                del pdata['COUNT'][0]
                del pdata['ERROR'][0]
                del pdata['FLUX'][0]
                del pdata['FERROR'][0]
                del pdata['CONTAM'][0]
            else:
                dec=0
