"""
$Revision: 1.6 $ $Date: 2010-08-25 06:47:17 $
Author: Martin Kuemmel
Affiliation: Space Telescope - European Coordinating Facility
"""
__author__ = "Martin Kuemmel <mkuemmel@eso.org>"
__date__ = "$Date: 2010-08-25 06:47:17 $"
__version__ = "$Revision: 1.6 $"
___credits__ = """This software was developed by the ACS group of the Space Telescope -
European Coordinating Facility (ST-ECF). The ST-ECF is a department jointly
run by the European Space Agency and the European Southern Observatory.
It is located at the ESO headquarters at Garching near Munich. The ST-ECF
staff supports the European astronomical community in exploiting the research
opportunities provided by the earth-orbiting Hubble Space Telescope.
"""

import string, os

def commlineparser(arglist):
    """
    Input:
        arglist - the command line given by the user

    Return:
        options - a dictionary with the user settings
                  stored as values for corresonding
                  keywords

    Description:
        The command line is analyzed for the existence
        of certain keywords. In case that a keyword
        is identified the next item is taken as the
        keyvalue. The keyvalue is stored in a dictionary
        with a corresponding key. Non-free entries (yes/no)
        are converted to upper case. It is determined whether
        some items of the command line could not be resolved.
        in that case a special key/value pair is set.
        This later causes a corresponding error.
        If no items are given in the command line this
        is interpreted as a help request.
    """
    options = {}
    argv_copy = []

# are there any items
    if len(arglist) == 1:
        options["HELP"] = 'yes'

# make a copy for the completeness check
    for arg in arglist:
        argv_copy.append(arg)

    for index in range(len(arglist)):
        if arglist[index] == "--parfile":
            options["PARFILE"] = arglist[index+1]
            argv_copy[index]   = None
            argv_copy[index+1] = None
        if arglist[index] == "--pagename":
            options["PAGENAME"] = arglist[index+1]
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--direct_image":
            options["DIRIMAGE"] = arglist[index+1]
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--stamp_image":
            options["STPIMAGE"] = arglist[index+1]
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--spectr_table":
            options["SPCTABLE"] = arglist[index+1]
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--sextract_cat":
            options["SEXTRCAT"] = arglist[index+1]
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--page_header":
            options["PAGEHEAD"] = arglist[index+1]
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--outputpath":
            options["OUTPPATH"] = arglist[index+1]
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--page_title":
            options["PAGETITL"] = arglist[index+1]
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--start_object":
            options["STARTOBJ"] = int(arglist[index+1])
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--number_objects":
            options["NUMBEOBJ"] = int(arglist[index+1])
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--pagesize":
            options["PAGESIZE"] = int(arglist[index+1])
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--spectrum_width":
            options["SPECWIDT"] = int(arglist[index+1])
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--directim_width":
            options["DIMWIDTH"] = int(arglist[index+1])
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--lambda_range":
            options["WAVERANG"] = arglist[index+1]
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--sort_column":
            options["SORTCOLU"] = string.upper(arglist[index+1])
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--error_bars":
            options["ERRORBAR"] = string.upper(arglist[index+1])
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--only_html":
            options["ONLYHTML"] = string.upper(arglist[index+1])
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--mark_sources":
            options["MARKSOUR"] = string.upper(arglist[index+1])
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--overview":
            options["OVERVIEW"] = string.upper(arglist[index+1])
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--data_out":
            options["DATAOUTP"] = string.upper(arglist[index+1])
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--mag_lambda":
            options["MAGLAMBD"] = int(float(arglist[index+1]))
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--css_file":
            options["CSSFILEU"] = arglist[index+1]
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--sel_file":
            options["SELEFILE"] = arglist[index+1]
            argv_copy[index]   = None
            argv_copy[index+1] = None
        elif arglist[index] == "--help":
            options["HELP"] = 'yes'
            argv_copy[index]   = None

# search for unknown settings
    unknown = 0
    index   = -1
    for arg in argv_copy:
        if arg != None:
            unknown = unknown + 1
            if index == -1 and argv_copy.index(arg) != 0:
                index = argv_copy.index(arg)
# if there are unknown settings, set the corresponding keyword
    if unknown > 1:
        options["UNKNOWN"] = argv_copy[index]

    return options

def makesettings(commline):
    """
    Input:
        commline   - dictionary  with user settings from
                     the command line

    Return:
        allsetting - dictionary with all possible
                     settings

    Description:
        The procedure puts together the settings from the
        parameter and from the command line and supplements
        non given values with default values if possible.
        Command line settings overwrite the parameterfile
        settings such that the user has full flexibility
        in each run. If for a given was neither done
        in the parameter file nor in the command line,
        a resonable default value is set.
    """
    allsettings = {}

# look if a parfile has been given
# load those settings into the dictionary.
# write then all command line settings in the dictionary
    if commline.has_key('PARFILE'):
        if os.access(commline['PARFILE'], os.R_OK) == 1:
            allsettings = analyzeparfile(commline['PARFILE'])
            allsettings.update(commline)
            for index in range(len(commline)):
                onekey = commline.popitem()
                if not allsettings.has_key(onekey[0]):
                    allsettings[onekey[0]] = onekey[1]
        else:
            allsettings['ERROR'] = 'NOACCESS'
    else:
        allsettings = commline.copy()


#  fill in resonable defaults
    if not allsettings.has_key('PAGENAME'):
        allsettings['PAGENAME'] = 'aXe2web'
    if not allsettings.has_key('PAGETITL'):
        allsettings['PAGETITL'] = 'aXe2web'
    if not allsettings.has_key('PAGEHEAD'):
        if allsettings.has_key('SEXTRCAT') and \
           len(string.strip(str(allsettings['SEXTRCAT']).split('.cat')[0])) > 0:
            allsettings['PAGEHEAD'] = \
                                    string.strip(str(allsettings['SEXTRCAT']).split('.cat')[0])
        else:
            allsettings['PAGETITL'] = 'aXe2web'
    else:
        if  allsettings['PAGEHEAD'][0] == '"' and \
        allsettings['PAGEHEAD'][len(allsettings['PAGEHEAD'])-1] == '"':
            headstring = allsettings['PAGEHEAD']
            headstring = headstring[1: len(headstring)-1]
            allsettings['PAGEHEAD'] = headstring
    if not allsettings.has_key('STARTOBJ'):
        allsettings['STARTOBJ'] = 1
    if not allsettings.has_key('NUMBEOBJ'):
        allsettings['NUMBEOBJ'] = -1
    if not allsettings.has_key('OUTPPATH'):
        allsettings['OUTPPATH'] = 'outhtml'
    if not allsettings.has_key('PAGESIZE'):
        allsettings['PAGESIZE'] = 50
    if not allsettings.has_key('SPECWIDT'):
        allsettings['SPECWIDT'] = 162
    if not allsettings.has_key('DIMWIDTH'):
        allsettings['DIMWIDTH'] = 50
    if not allsettings.has_key('WAVERANG'):
        allsettings['WAVERANG'] = '5500,10500'
    if not allsettings.has_key('SORTCOLU'):
        allsettings['SORTCOLU'] = None
        allsettings['ASCENDIN'] = 'YES'
    else:
        if allsettings['SORTCOLU'][len(allsettings['SORTCOLU'])-1] == '-':
            allsettings['ASCENDIN'] = 'NO'
            allsettings['SORTCOLU'] = allsettings['SORTCOLU'][0:len(allsettings['SORTCOLU'])-1]
        elif allsettings['SORTCOLU'][len(allsettings['SORTCOLU'])-1] == '+':
            allsettings['ASCENDIN'] = 'YES'
            allsettings['SORTCOLU'] = allsettings['SORTCOLU'][0:len(allsettings['SORTCOLU'])-1]
        else:
            allsettings['ASCENDIN'] = 'YES'
    if not allsettings.has_key('ERRORBAR'):
        allsettings['ERRORBAR'] = 'YES'
    if not allsettings.has_key('ONLYHTML'):
        allsettings['ONLYHTML'] = 'NO'
    if not allsettings.has_key('MARKSOUR'):
        allsettings['MARKSOUR'] = 'YES'
    if not allsettings.has_key('OVERVIEW'):
        allsettings['OVERVIEW'] = 'YES'
    if not allsettings.has_key('DATAOUTP'):
        allsettings['DATAOUTP'] = 'NO'
    if not allsettings.has_key('MAGLAMBD'):
        allsettings['MAGLAMBD'] = 800.0
    if not allsettings.has_key('CSSFILEU'):
        allsettings['CSSFILEU'] = None
    if not allsettings.has_key('SELEFILE'):
        allsettings['SELEFILE'] = None

    return allsettings


def analyzeparfile(filename):
    """
    Input:
        filename    - the name of the parameter file

    Return:
        filesettings - dictionary with the settings given
                       in the parameter file

    Description:
        Each line of the parameter file is read in.
        Comment lines and empty lines are ignored.
        Lines with settings are split up by the
        routine 'analyzeline' into keyword and
        keyvalue. The keyword/keyvalue pairs are
        then stored in the return dictionary.
        In case that a unknown setting is found
        the keyword 'UNKNOWN' is filled.
    """
    filesettings = {}

    pfile = open(filename, 'r')

    plines = pfile.readlines()

    for oneline in plines:
        if len(string.strip(oneline)) > 1 and string.strip(oneline)[0] != '#':
            optvalpair = analyzeline(oneline)
            if optvalpair[0] == 'pagename':
                filesettings["PAGENAME"] = str(optvalpair[1])
            elif optvalpair[0] == 'direct_image':
                filesettings["DIRIMAGE"] = str(optvalpair[1])
            elif optvalpair[0] == 'stamp_image':
                filesettings["STPIMAGE"] = str(optvalpair[1])
            elif optvalpair[0] == 'spectr_table':
                filesettings["SPCTABLE"] = str(optvalpair[1])
            elif optvalpair[0] == 'sextract_cat':
                filesettings["SEXTRCAT"] = str(optvalpair[1])
            elif optvalpair[0] == 'page_header':
                filesettings["PAGEHEAD"] = str(optvalpair[1])
            elif optvalpair[0] == 'outputpath':
                filesettings["OUTPPATH"] = str(optvalpair[1])
            elif optvalpair[0] == 'page_title':
                filesettings["PAGETITL"] = str(optvalpair[1])
            elif optvalpair[0] == 'start_object':
                filesettings["STARTOBJ"] = int(optvalpair[1])
            elif optvalpair[0] == 'number_objects':
                filesettings["NUMBEOBJ"] = int(optvalpair[1])
            elif optvalpair[0] == 'pagesize':
                filesettings["PAGESIZE"] = int(optvalpair[1])
            elif optvalpair[0] == 'spectrum_width':
                filesettings["SPECWIDT"] = int(optvalpair[1])
            elif optvalpair[0] == 'directim_width':
                filesettings["DIMWIDTH"] = int(optvalpair[1])
            elif optvalpair[0] == 'lambda_range':
                filesettings["WAVERANG"] = str(optvalpair[1])
            elif optvalpair[0] == 'sort_column':
                filesettings["SORTCOLU"] = string.upper(str(optvalpair[1]))
            elif optvalpair[0] == 'error_bars':
                filesettings["ERRORBAR"] = string.upper(str(optvalpair[1]))
            elif optvalpair[0] == 'only_html':
                filesettings["ONLYHTML"] = string.upper(str(optvalpair[1]))
            elif optvalpair[0] == 'mark_sources':
                filesettings["MARKSOUR"] = string.upper(str(optvalpair[1]))
            elif optvalpair[0] == 'overview':
                filesettings["OVERVIEW"] = string.upper(str(optvalpair[1]))
            elif optvalpair[0] == 'data_out':
                filesettings["DATAOUTP"] = string.upper(str(optvalpair[1]))
            elif optvalpair[0] == 'mag_lambda':
                filesettings["MAGLAMBD"] = int(float(optvalpair[1]))
            elif optvalpair[0] == 'css_file':
                filesettings["CSSFILEU"] = str(optvalpair[1])
            elif optvalpair[0] == 'sel_file':
                filesettings["SELEFILE"] = str(optvalpair[1])
            else:
                filesettings["UNKNOWNF"] = str(optvalpair[0]) + ' ' + str(optvalpair[1])

    return filesettings

def analyzeline(oneline):
    """
    Input:
        oneline  - a line from the parameter file
    Return:
        option   - a list with option key (string) and
                   optinvalue

    Description:
        Analyzes a line in the parameter file and splits it up
        at a '=' sign inot option key and optionvalue.
        fo a better access the key/value pair is stored
        and given back in a list.
    """
    option = []
    equpos = string.find(oneline, '=')
    if equpos > -1:
        optstring = string.strip(oneline.split('=')[0])
        optvalue  = string.strip(oneline.split('=')[1])
        commpos = string.find(optvalue, '#')
        if commpos > -1:
            optvalue = string.strip(optvalue[0:commpos-1])
        else:
            optvalue = string.strip(optvalue)
    else:
        optstring = 'UNKNOWNF'
        commpos = string.find(oneline, '#')
        if commpos > -1:
            optvalue = string.strip(oneline[0:commpos-1])
        else:
            optvalue = string.strip(oneline)

    option.append(optstring)
    option.append(optvalue)

    return option

def checksettings(settings):
    """
    Input:
        settings - all settings

    Return:
        an error code

    Description:
        The module checks whether all the required settings
        were done aorrectly. There are only few settings which have
        to be given. Those depend on other settings, e.g
        no direct image has to be given in case
        '--only_html YES'. Moreover the check for unknown
        settings is done. It is checked whether at the
        according settings only 'YES' or 'NO' is given.
        Case sensitive errorcodes are returned incase that
        wrong or missing settings prevent further execution
        of the program.
    """
    # here you define the name of the image directory
    # relative to the output directory!!
    r_imdir   = 'images'
    r_datdir  = 'data'
    r_fitsdir = 'fits'

    if settings.has_key('HELP'):
        return -1

    if settings.has_key("UNKNOWN"):
        print settings['UNKNOWN']
        return 6
    if settings.has_key("UNKNOWNF"):
        print 'Setting in the parameter file: ', settings['UNKNOWNF']
        return 7
    if not settings.has_key("SEXTRCAT"):
        return 2
    elif os.access(settings["SEXTRCAT"], os.R_OK) != 1:
        return 21

    if not settings.has_key("DIRIMAGE") and settings["ONLYHTML"] != 'YES':
        return 3
    elif settings.has_key("DIRIMAGE") and os.access(settings["DIRIMAGE"], os.R_OK) != 1:
        dirimage = str(settings["DIRIMAGE"])
        if string.find(dirimage, '[') != -1 and  string.find(dirimage, ']') != -1:
            dirimage = dirimage[0:int(string.find(dirimage, '['))]
            if os.access(dirimage, os.R_OK) != 1:
                return 31
            else:
                print 'Found ', dirimage, '!'
        else:
            return 31

    if not settings.has_key("STPIMAGE") and settings["ONLYHTML"] != 'YES':
        return 4
    elif settings.has_key("STPIMAGE") and os.access(settings["STPIMAGE"], os.R_OK) != 1:
        return 41
    elif settings["CSSFILEU"] != None and os.access(settings["CSSFILEU"], os.R_OK) != 1:
        return 42
    elif settings["SELEFILE"] != None and os.access(settings["SELEFILE"], os.R_OK) != 1:
        return 43
    if not settings.has_key("SPCTABLE") and settings["ONLYHTML"] != 'YES':
        return 5
    elif settings.has_key("SPCTABLE") and os.access(settings["SPCTABLE"], os.R_OK) != 1:
        return 51
    if settings.has_key("UNKNOWN"):
        print 'Setting in command line: ', settings['UNKNOWN']
        return 6
    if settings.has_key("UNKNOWNF"):
        print 'Setting in the parameter file: ', settings['UNKNOWNF']
        return 7
    if not (settings['ERRORBAR'] == 'YES' or settings['ERRORBAR'] == 'NO'):
        return 81
    if not (settings['ONLYHTML'] == 'YES' or settings['ONLYHTML'] == 'NO'):
        return 82
    if not (settings['OVERVIEW'] == 'YES' or settings['OVERVIEW'] == 'NO'):
        return 83
    if not (settings['MARKSOUR'] == 'YES' or settings['MARKSOUR'] == 'NO'):
        return 85
    pname = settings['OUTPPATH']
    pname = analyzepathname(pname)
    if pname == 91 or pname == 92 or pname == 93:
        return pname
    else:
        settings['OUTPPATH'] = pname

    #
    # create the image directory
    #
    pname =  settings['OUTPPATH'] + '/' + r_imdir
    pname = analyzepathname(pname)
    if pname == 91 or pname == 92 or pname == 93:
        return pname
    else:
        settings['IMAGPATH'] = r_imdir

    # comment these lines if the
    # direct fits stamps should NOT be saved
    #
    # create the fits directory
    #
    pname =  settings['OUTPPATH'] + '/' + r_fitsdir
    pname = analyzepathname(pname)
    if pname == 91 or pname == 92 or pname == 93:
        return pname
    else:
        settings['FITSPATH'] = r_fitsdir

    #
    # create the data directory
    #
    if settings['DATAOUTP'] == 'YES':
        pname =  settings['OUTPPATH'] + '/' + r_datdir
        pname = analyzepathname(pname)
        if pname == 91 or pname == 92 or pname == 93:
            return pname
        else:
            settings['DATAPATH'] = r_datdir
    else:
        settings['DATAPATH'] = None


    if not settings.has_key("WAVERANG"):
        return 110
    else:
        wrange = string.strip(str(settings['WAVERANG']))
        if string.find(wrange, ',') < 0:
            return 111
        else:
            if wrange[0] == ',' or string.split(wrange, ',')[0] == '*':
                settings['WAVLOWER'] = 5500
            else:
#               settings['WAVLOWER'] = int(string.split(wrange, ',')[0])
                settings['WAVLOWER'] = float(string.split(wrange, ',')[0])
            if wrange[len(wrange)-1] == ',' or string.split(wrange, ',')[1] == '*':
                settings['WAVUPPER'] = 10500
            else:
#               settings['WAVUPPER'] = int(string.split(wrange, ',')[1])
                settings['WAVUPPER'] = float(string.split(wrange, ',')[1])
            if settings['WAVLOWER'] > settings['WAVUPPER']:
                tmp =  settings['WAVUPPER']
                settings['WAVUPPER'] = settings['WAVLOWER']
                settings['WAVLOWER'] = tmp
                print ''
                print 'The lower cut was larger than the higher wavelength cut.'
                print 'Upper and lower limits therefore are exchanged.'
                print ''
#           if settings['WAVLOWER'] < 1000:
#               print ''
#               print 'Lower wavelength cut is unreasonable (<1000 AA).'
#               print 'Setting lower cut to 1000AA'
#               print ''
#               settings['WAVUPPER'] = 1000
#           if settings['WAVUPPER'] > 30000:
#               print ''
#               print 'Upper wavelength cut is unreasonable (>30000 AA).'
#               print 'Setting upper cut to 30000AA'
#               print ''
#               settings['WAVUPPER'] = 30000
            if not settings.has_key('WAVLOWER'):
                return 112
            if not settings.has_key('WAVUPPER'):
                return 113
            if settings.has_key('WAVUPPER') and settings.has_key('WAVLOWER'):
                del settings['WAVERANG']

    return 0


def analyzepathname(pname):
    """
    Input:
        pname - a directory name

    Return:
        pname - the absolute pathname for pname
                or an error code

    Description:
        Checks whether the directory for thew output exists and
        whether it is usefull. If not the directory is created.
        If the program can proceed without problems, the absolut
        pathname of pname is returned, otherwise a
        case sensitive errorcode is given back.
    """
    if os.path.exists(pname) ==1:
        if not (os.path.isdir(pname) == 1 and os.access(pname, os.W_OK) ==1):
            return 91
    else:
        try:
            os.mkdir(pname)
        except OSError:
            return 92
        if not (os.path.isdir(pname) == 1 and os.access(pname, os.W_OK) ==1):
            return 93
    pname = os.path.abspath(pname)

    return pname
