"""
$Revision: 1.3 $ $Date: 2010-08-25 06:47:17 $
Author: Martin Kuemmel
Affiliation: Space Telescope - European Coordinating Facility
"""
__author__ = "Martin Kuemmel <mkuemmel@eso.org>"
__date__ = "$Date: 2010-08-25 06:47:17 $"
__version__ = "$Revision: 1.3 $"
___credits__ = """This software was developed by the ACS group of the Space Telescope -
European Coordinating Facility (ST-ECF). The ST-ECF is a department jointly
run by the European Space Agency and the European Southern Observatory.
It is located at the ESO headquarters at Garching near Munich. The ST-ECF
staff supports the European astronomical community in exploiting the research
opportunities provided by the earth-orbiting Hubble Space Telescope.
"""
import os
import os.path
import sys
import pyfits

class FitsSlicer(object):
    """
    Class for slicing  fits files
    """
    def __init__(self, fits_name, name_string=None, name_map=None, verbose=True):
        """
        Initializes the class
        """
        # store all input
        self.fits_name   = fits_name
        self.name_string = name_string
        self.name_map    = name_map.copy()
        self.verbose     = verbose
        
        # make sure the input is correct
        self._initialize(fits_name, name_string, name_map)

    def _initialize(self, fits_name, name_string, name_map):
        """
        Check whether the input is correct
        """
        # make sure the fits file is there
        if not os.path.isfile(fits_name):
            err_msg = 'The fits file: %s does not exist!' % fits_name
            raise Exception(err_msg)
        
        # check that you can assign name    
        if not name_string and not name_map:
            # error and out
            err_msg = "Please specify either a naming scheme OR a name map!"
            raise Exception(err_msg)

        # make sure the naming scheme works
        if name_string:
            try:
                # make an example
                new_name = name_string % 'BEAM_1A'
            except TypeError:
                # error and out
                err_msg = 'The naming scheme "%s" does not work!' % name_string
                raise Exception(err_msg)

    def _get_new_name(self, fits_ext):
        """
        """
        # give some defaults
        go_on    = True
        new_name = None

        # make sure there is data
        if fits_ext.data == None:
            # return the results
            return new_name, go_on

        # make sure the extension name is set
        if not fits_ext.header.has_key('EXTNAME'):
            # return the results
            return new_name, go_on

        # set the extension name
        ext_name = fits_ext.header['EXTNAME']

        # check for a name map
        if self.name_map != None:

            # check whether current extension is on the map
            if ext_name in self.name_map:

                # set the new name and delete its entry
                new_name = self.name_map[ext_name]
                del self.name_map[ext_name]

            else:
                # give back the NULL results
                return new_name, go_on 
        else:
            # form the new name with the string
            new_name = self.name_string % ext_name

        # make sure there is still something to look for
        if self.name_map != None and len(self.name_map) < 1:
            go_on = False
    
        # return the results
        return new_name, go_on

    def _make_new_fits(self, current_ext, new_name):
        """
        """
        # make a new primary HDU
        prim_hdu = pyfits.PrimaryHDU()

        # make a new extension list
        thdulist = pyfits.HDUList([prim_hdu, current_ext])
        
        # delete previous versions
        if os.path.isfile(new_name):
            os.unlink(new_name)

        # write to a file
        thdulist.writeto(new_name)

    def run(self):
        """
        Slice the fits
        """
        # give a brief feedback
        print 'Start slicing fits-file: ', self.fits_name, ' ...',
        sys.stdout.flush()

        # open the fits file
        fits = pyfits.open(self.fits_name, 'readonly')

        # initialize some stuff
        n_ext = len(fits)
        ext_index = 0
        go_on = True
        
        # continue as long as possible
        while ext_index < n_ext and go_on:

            # define the current extension;
            # report if desired
            current_ext = fits[ext_index]
            if self.verbose:
                print "Checking extension No.: ", ext_index

            # check whether the current extension
            # needs top be sliced; 
            # get the name of the slice 
            new_name, go_on = self._get_new_name(current_ext)

            # check for a name
            if new_name:
                
                # write extension to this name
                self._make_new_fits(current_ext, new_name)

                # give feedback if desired
                if self.verbose:
                    message = "File: %s extension No.: %i written as: %s!" % (self.fits_name, ext_index, new_name)
                    print message
            # enhance the fits index
            ext_index += 1

        # close the fits file
        fits.close()
        
        # give a brief feedback
        print ' Done.'
