"""
$Revision: 1.3 $ $Date: 2010-08-25 06:47:17 $
Author: Martin Kuemmel
Affiliation: Space Telescope - European Coordinating Facility
"""
__author__ = "Martin Kuemmel <mkuemmel@eso.org>"
__date__ = "$Date: 2010-08-25 06:47:17 $"
__version__ = "$Revision: 1.3 $"
___credits__ = """This software was developed by the ACS group of the Space Telescope -
European Coordinating Facility (ST-ECF). The ST-ECF is a department jointly
run by the European Space Agency and the European Southern Observatory.
It is located at the ESO headquarters at Garching near Munich. The ST-ECF
staff supports the European astronomical community in exploiting the research
opportunities provided by the earth-orbiting Hubble Space Telescope.
"""

import sys, os, string,math
#
#
#----------------------------------------------------------------------
#
#----------------------------------------------------------------------
def computestat(vector):
    """
    Input:
        vector   - list with numerical values

    Return:
        statvals - dictionary

    Description:
        Computes basic statistic from the values stored in the
        input vector. The statistical quantities are stored and
        returned in a dictionary. The keywords and
        their meaning is:

        - min: the minimum value
        - max: the maximum value
        - sum: the sum of all values
        - ave: the average of all values
        - std: the standard deviation

        In case that the incoming vector has length 0
        all dictionary entries are set to 0.0.

    """
    statvals = dict()
    sum = 0.0
    std = 0.0

#  go out with 0.0 if nothing comes in
    if len(vector)==0:
        statvals["min"] = 0.0
        statvals["max"] = 0.0
        statvals["sum"] = 0.0
        statvals["ave"] = 0.0
        statvals["std"] = 0.0
        return statvals

# a first pass to find min and max and to
# get the sum and average
    min = vector[0]
    max = vector[0]
    for index in range(len(vector)):
        if vector[index] > max:
            max =  vector[index]
        if vector[index] < min:
            min =  vector[index]
        sum = sum + vector[index]
    ave = sum / len(vector)

# second pass for the standard deviation
    for index in range(len(vector)):
        std = std + (vector[index] - ave)**2
    if len(vector) > 1:
        std = std /(float(len(vector))-1.0)
    else:
        std = std /float(len(vector))

    statvals["min"] = min
    statvals["max"] = max
    statvals["sum"] = sum
    statvals["ave"] = ave
    statvals["std"] = std

    return statvals

#----------------------------------------------------------------------
#
#----------------------------------------------------------------------
def rescale_int(vector, min, max):

    scalevector = []
    oldstat  = computestat(vector)
    oldrange = oldstat["max"]- oldstat["min"]
    newrange = max - min
    scale = newrange/oldrange
    offset = min - oldstat["min"]
    for index in range(len(vector)):
        scalevector.append((vector[index]+offset)*scale)
    return scalevector

def rescale_int(vector, min_in, max_in, min_out, max_out):
    """
    Input:
        vector      - list with numerical values
        min_in      - minimum of input range
        max_in      - maximum of the input range
        min_out     - minimum of scaled range
        max_out     - maximum of the scaled range

    Return:
        scalevector - list with the scaled values of vector

    Description:
        Rescales the values in the input vector. The scaling
        is done such that the value range [min_in:max_in]
        is linearly scaled to [min_out:max_out].
        In case that the incoming vector has only constant
        values, all values are scaled to the middle of
        the output range.
        Scaled values which ly outside the outputrange
        are cut to min_out or max_out, respectively.
    """

    scalevector = []

# determine the ranges
    oldrange = max_in - min_in
    newrange = max_out - min_out

# see whether there are only constant input values
# determine scale and offset
    if oldrange != 0.00:
        scale = newrange/oldrange
        offset = min_out - min_in
    else:
        scale = 1.0
        offset = min_out - min_in + 0.5* newrange

# apply scale and offset, cut the outlyers
    for index in range(len(vector)):
#       scalevector.append((vector[index]+offset)*scale)
        value = (vector[index]+offset)*scale
        if value < min_out:
            value = min_out
        elif value > max_out:
            value = max_out
        scalevector.append(value)

    return scalevector

def getscales(vector, min, max):
    """
    Input:
        vector    - list with numerical values
        min       - minimum value
        max       - maximum value

    Return:
        scaleoffs - scaling factor and scaling offset

    Description:
        Derives the scaling parameters scaling factor
        and scaling factor given an input vector
        and the output range.
        Small helping function which probaby
        is no longer use.
    """
    scaleoffs = []

    oldstat  = computestat(vector)
    oldrange = oldstat["max"]- oldstat["min"]
    newrange = max - min
    scaleoffs.append(newrange/oldrange)
    scaleoffs.append((min - oldstat["min"])/scaleoffs[0])

    return scaleoffs

def sortvector(vector):
    """
    Input:
        vector    - list with numerical values

    Return:
        Formally there is no return value, however
        the input vector itself is changed.
        Afterwards the calling program has those changed
        values, too.

    Description:
        A formal frontend to the intrinsic list function
        'sort'.
    """
    vector.sort()

def getpercentiles(vector, pc1, pc2):
    """
    Input:
        vector    - list with numerical values
        pc1       - one percentile value; unit [%]
        pc2       - the second percentile value; unit [%]

    Return:
        percentiles - a list with the two percentiles
                      derived from the input vector

    Description:
        derives the percentiles for the values in
        the incoming vector. E.g. the call
        'getpercentiles(invec, 10, 90)' returns the
        val1 and val2. 90% of the entries in 'invec'
        are larger than val1, and 10% of the entries are
        larer than val2.
        I did NOT aim for mathematical correctness.
        No interpolation between to vectorentries is
        done, and perhaps the result is not
        always correct. The routine only aims to
        give reasonable cut values for images
        indpendent on their dynamical scale.
    """
    import numpy

    tmp = []
    percentiles = []
    veclen = len(vector)
    for index in range(veclen):
        # dirty trick to avoid NaN's
        if numpy.isnan(vector[index]):
            continue
        tmp.append(vector[index])
    tmp.sort()

    index_p1 = int(math.floor(pc1/100.0*float(veclen) + 0.5) - 1)
    if index_p1 > veclen-1:
        index_p1 = veclen-1

    elif index_p1 < 0:
        index_p1 = 0

    index_p2 = int(math.floor(pc2/100.0*float(veclen) + 0.5) - 1)
    if index_p2 > veclen-1:
        index_p2 = veclen-1

    elif index_p2 < 0:
        index_p2 = 0

    try:
        percentiles.append(tmp[index_p1])
        percentiles.append(tmp[index_p2])
    except:
        percentiles.append(0.0)
        percentiles.append(0.0)

    return percentiles

def radectostring(ra, dec):
    """
    Input:
       ra  - float value which represents right ascension in
             degrees (0--360)
       dec - float value which represents declination in
             degrees (-90 -- 90)

    Return:
       radec - a list with the string representation
               of the input coordinates

    Description:
       In cataloges right ascension and declination are often
       notated in units of degrees. However astronomers
       are used to look at them in the usual
       hh:mm:ss.s dd:mm:ss.s version. This function
       does the transformation from the degree-version
       to the 'astronomical' version. Between the subunits
       ':' are placed..
    """
    radec = []
    tmp1 = ra/15.0
    rah  = int(math.floor(tmp1))
    tmp2 = tmp1-float(rah)
    ram  = int(math.floor((tmp2*60)))
    tmp3 = tmp2*60 - float(ram)
    ras  = 60*tmp3
    radec.append('%3i:%02i:%05.2f'% (rah,ram,ras))
    if dec < 0:
        deg  = int(math.ceil(dec))
        tmp1 = -1.0*(dec-float(deg))
    else:
        deg  = int(math.floor(dec))
        tmp1 = dec-float(deg)
    dem  = int(math.floor((tmp1*60)))
    tmp2 = tmp1*60 - float(dem)
    des  = 60*tmp2
    radec.append('%3i:%02i:%04.1f'% (deg,dem,des))

    return radec

def  filtZEROminusINF(invec):
    outvec = []
    for data in invec:
        if math.fabs(data) > 1.0E-20 and data > -9.0E+05:
            outvec.append(data)

    return outvec

#----------------------------------------------------------------------
#
