"""
$Revision: 1.3 $ $Date: 2010-08-25 06:47:17 $
Author: Martin Kuemmel
Affiliation: Space Telescope - European Coordinating Facility
"""
__author__ = "Martin Kuemmel <mkuemmel@eso.org>"
__date__ = "$Date: 2010-08-25 06:47:17 $"
__version__ = "$Revision: 1.3 $"
___credits__ = """This software was developed by the ACS group of the Space Telescope -
European Coordinating Facility (ST-ECF). The ST-ECF is a department jointly
run by the European Space Agency and the European Southern Observatory.
It is located at the ESO headquarters at Garching near Munich. The ST-ECF
staff supports the European astronomical community in exploiting the research
opportunities provided by the earth-orbiting Hubble Space Telescope.
"""

#----------------------------------------------------------------------------
#
#
class GeneralFileName:
    """Class to derive the names for all type of filenames. The filenames
       are composed by at most  three items:
    1. The specific rootname
    2. A postfix to the rootname
    3. An extension to derive filenames for various filetypes

    In this variable class all thre components can be given at
    initialization as well as in corresponding set-methods.
    The full name as well as the name components can be derived
    by get-methods. For various naming conventions extension names
    can be derived on the fly (e.g. for the convention [BEAM_'num'A].
    """
    def __init__(self, rootname, postfix, filetype):
        """The rootname is combined with the postfix and filetype at
        initialization. All components are stored in case that later a
        one of them is renewed. new rootname is given."""
        self.nameroot = str(rootname)
        self.postfix  = str(postfix)
        self.filetype = str(filetype)
        self.beamnum  = ''
        self.fullname = self.nameroot + self.postfix + self.filetype
    def setpostfix(self, postfix):
        """Set a new name by giving a new postfix."""
        self.postfix = str(postfix)
        self.fullname = self.nameroot + self.postfix + self.filetype
    def getpostfix(self):
        """Return the actual postfix of the 'object'."""
        return self.postfix
    def setrootname(self, rootname):
        """Set a new rootname."""
        self.nameroot = rootname
        self.fullname = self.nameroot + self.postfix + self.filetype
    def getrootname(self):
        """Return the root name of the object."""
        return self.nameroot
    def setfiletype(self, filetype):
        """Set a new filetype name."""
        self.filetype = filetype
        self.fullname = self.nameroot + self.postfix + self.filetype
    def getfiletype(self):
        """Return the filetype name of the object."""
        return self.filetype
    def setbeamextname(self, beamnum):
        """Set the beam number."""
        self.beamnum = str(beamnum)
        self.beamext = '[BEAM_' + self.beamnum + 'A]'
        self.fullname = self.nameroot + self.postfix + self.filetype + self.beamext
    def getbeamnum(self):
        """Return the beam number of the object."""
        return self.beamnum
    def getname(self):
        """Return the actual name of the 'object'."""
        return self.fullname
    def getbeamextname(self, beamnum):
        """Return the name of the object plus beam extension according
        to 'fullname' + '[BEAM_'num'A]'"""
        self.beamnum = str(beamnum)
        typename = self.fullname + self.beamext
        return typename
    def getbeamextname(self):
        """Return the name of the object plus beam extension."""
        typename = self.fullname + '[BEAM_' + self.beamnum + 'A]'
        return typename
#
#----------------------------------------------------------------------------
    
#----------------------------------------------------------------------------
#
#
class HTMLPageName:
    """Class to derive filenames for the html pages. The filenames are
    composed by at most two items:
    1. The specific rootname given at initialization or by the
       corresponding set method.
    2. A postfix to the rootname whhich is composed by
       '_p'+ number. The number corresponds to the pagenumber
       in the current set of html pages.

    There are are only get and set methods to compose and get
    the different name-items  for the object.
    """
    def __init__(self, rootname, postfix):
        """The rootname is combined with the postfix at initialization.
        The postfix is stored in case that later a new rootname is given,
        and the a new name has to be built"""
        self.nameroot = str(rootname)
        if len(str(postfix)) > 0:
            self.postfix = '_p' + str(postfix)
        else:
            self.postfix = ''
        self.fullname = self.nameroot + self.postfix + '.html'
    def setpostfix(self, postfix):
        """Set a new name by giving a new postfix."""
        self.postfix = '_p' + str(postfix)
        self.fullname = self.nameroot + self.postfix + '.html'
    def getname(self):
        """Return the actual name of the 'object'."""
        return self.fullname
    def getpostfix(self):
        """Return the actual postfix of the 'object'."""
        return self.postfix
    def setrootname(self, rootname):
        """Set a new rootname (deprecated!!)."""
        self.nameroot = rootname
        self.fullname = self.nameroot + self.postfix + '.html'
    def getrootname(self):
        """Return the root name of the object."""
        return self.nameroot
#
#----------------------------------------------------------------------------


#----------------------------------------------------------------------------
#
#
class CountsPlotName:
    """Class to derive filenames for the graphics of the spectra in
    counts/sec. The filenames are composed bye four items:
    1. The specific rootname 'pltcts'
    2. The flavour 'b' or 'n' to indicate a plot with or without
       errorbars
    3. A postfix to the rootname, mostly the object number
    4. An extension to derive filenames for various filetypes

    There are are only get and set methods to compose and get
    the full names for the various filetypes. Although
    there is a method to change the rootname, it is not recommended
    to do so since the rootname is supposed to be proprietary
    to the class.
    """
#----------------------------------------------------------------------------
#
    nameroot = 'pltcts'
#
#----------------------------------------------------------------------------

    def __init__(self, postfix, ebar):
        """The rootname is combined with the postfix at initialization.
        The postfix is stored in case that later a new rootname is given,
        and the a new name has to be built"""
        self.postfix = str(postfix)
        if ebar == 'YES':
            self.flavour = 'b'
        else:
            self.flavour = 'n'
        self.fullname = self.nameroot + self.flavour + str(postfix)

        self.defaultimname = self.getpngname()

    def getdefimname(self):
        return self.defaultimname
    def setname(self, postfix):
        """Set a new name by giving a new postfix."""
        self.postfix = str(postfix)
        self.fullname = self.nameroot + self.flavour + str(postfix)
    def getname(self):
        """Return the actual name of the 'object'."""
        return self.fullname
    def getpostfix(self):
        """Return the actual postfix of the 'object'."""
        return self.postfix
    def setflavour(self, flavour):
        """Set a new flavour (deprecated!!)."""
        if flavour == 'YES':
            self.flavour = 'b'
        else:
            self.flavour = 'n'
        self.flavour = flavour
        self.fullname = self.nameroot + self.flavour + self.postfix
    def getflavour(self):
        """Return the flavour of the object."""
        return self.flavour
    def setrootname(self, rootname):
        """Set a new rootname (deprecated!!)."""
        self.nameroot = rootname
        self.fullname = self.nameroot + self.flavour + self.postfix
    def getrootname(self):
        """Return the root name of the object."""
        return self.nameroot
    def getjpgname(self):
        """Return the name of the object as filetype jpeg-image."""
        typename = self.fullname + '.jpg'
        return typename
    def getpngname(self):
        """Return the name of the object as filetype png-image."""
        typename = self.fullname + '.png'
        return typename
    def getpbmname(self):
        """Return the name of the object as filetype pbm-image."""
        typename = self.fullname + '.pbm'
        return typename
    def gettifname(self):
        """Return the name of the object as filetype tif-image."""
        typename = self.fullname + '.tif'
        return typename
    def getgifname(self):
        """Return the name of the object as filetype gif-image."""
        typename = self.fullname + '.gif'
        return typename
#
#----------------------------------------------------------------------------
    
#----------------------------------------------------------------------------
#
#
class FluxPlotName:
    """Class to derive filenames for the graphics of the spectra in
    physical units. The filenames are composed bye four items:
    1. The specific rootname 'pltflx'
    2. The flavour 'b' or 'n' to indicate a plot with or without
       errorbars
    3. A postfix to the rootname, mostly the object number
    4. An extension to derive filenames for various filetypes

    There are are only get and set methods to compose and get
    the full names for the various filetypes. Although
    there is a method to change the rootname, it is not recommended
    to do so since the rootname is supposed to be proprietary
    to the class.
    """
#----------------------------------------------------------------------------
#
    nameroot = 'pltflx'
#
#----------------------------------------------------------------------------

    def __init__(self, postfix, ebar):
        """The rootname is combined with the postfix at initialization.
        The postfix is stored in case that later a new rootname is given,
        and the a new name has to be built"""
        self.postfix = str(postfix)
        if ebar == 'YES':
            self.flavour = 'b'
        else:
            self.flavour = 'n'
        self.fullname = self.nameroot + self.flavour + str(postfix)

        self.defaultimname = self.getpngname()

    def getdefimname(self):
        return self.defaultimname
    def setname(self, postfix):
        """Set a new name by giving a new postfix."""
        self.postfix = str(postfix)
        self.fullname = self.nameroot + self.flavour + str(postfix)
    def getname(self):
        """Return the actual name of the 'object'."""
        return self.fullname
    def getpostfix(self):
        """Return the actual postfix of the 'object'."""
        return self.postfix
    def setflavour(self, flavour):
        """Set a new flavour (deprecated!!)."""
        if flavour == 'YES':
            self.flavour = 'b'
        else:
            self.flavour = 'n'
        self.flavour = flavour
        self.fullname = self.nameroot + self.flavour + self.postfix
    def getflavour(self):
        """Return the flavour of the object."""
        return self.flavour
    def setrootname(self, rootname):
        """Set a new rootname (deprecated!!)."""
        self.nameroot = rootname
        self.fullname = self.nameroot + str(postfix)
    def getrootname(self):
        """Return the root name of the object."""
        return self.nameroot
    def getjpgname(self):
        """Return the name of the object as filetype jpeg-image."""
        typename = self.fullname + '.jpg'
        return typename
    def getpngname(self):
        """Return the name of the object as filetype png-image."""
        typename = self.fullname + '.png'
        return typename
    def getpbmname(self):
        """Return the name of the object as filetype pbm-image."""
        typename = self.fullname + '.pbm'
        return typename
    def gettifname(self):
        """Return the name of the object as filetype tif-image."""
        typename = self.fullname + '.tif'
        return typename
    def getgifname(self):
        """Return the name of the object as filetype gif-image."""
        typename = self.fullname + '.gif'
        return typename
#
#----------------------------------------------------------------------------
    
#----------------------------------------------------------------------------
#
#
class DirectStampName:
    """Class to derive filenames for the direct image stamp files
    The filenames are composed bye three items:
    1. The specific rootname 'stpdim'
    2. A postfix to the rootname, mostly the object number
    3. An extension to derive filenames for various filetypes

    There are are only get and set methods to compose and get
    the full names for the various filetypes. Although
    there is a method to change the rootname, it is not recommended
    to do so since the rootname is supposed to be proprietary
    to the class.
    """
#----------------------------------------------------------------------------
#
    nameroot = 'stampdim'
#
#----------------------------------------------------------------------------

    def __init__(self, postfix, mark):
        """The rootname is combined with the postfix at initialization.
        The postfix is stored in case that later a new rootname is given,
        and the a new name has to be built"""
        self.postfix = str(postfix)
        if mark == 'YES':
            self.flavour = 'm'
        else:
            self.flavour = 'n'
        self.fullname = self.nameroot + self.flavour + str(postfix)

        # set the image name
        self.defaultimname = self.getpngname()

    def getdefimname(self):
        return self.defaultimname
    def setname(self, postfix):
        """Set a new name by giving a new postfix."""
        self.postfix = str(postfix)
        self.fullname = self.nameroot + self.flavour + str(postfix)
    def getname(self):
        """Return the actual name of the 'object'."""
        return self.fullname
    def getpostfix(self):
        """Return the actual postfix of the 'object'."""
        return self.postfix
    def setrootname(self, rootname):
        """Set a new rootname (deprecated!!)."""
        self.nameroot = rootname
        self.fullname = self.nameroot + self.flavour + self.postfix
    def getrootname(self):
        """Return the root name of the object."""
        return self.nameroot
    def getjpgname(self):
        """Return the name of the object as filetype jpeg-image."""
        typename = self.fullname + '.jpg'
        return typename
    def getpngname(self):
        """Return the name of the object as filetype png-image."""
        typename = self.fullname + '.png'
        return typename
    def getpbmname(self):
        """Return the name of the object as filetype png-image."""
        typename = self.fullname + '.pbm'
        return typename
    def gettifname(self):
        """Return the name of the object as filetype tif-image."""
        typename = self.fullname + '.tif'
        return typename
    def getgifname(self):
        """Return the name of the object as filetype gif-image."""
        typename = self.fullname + '.gif'
        return typename
    def getfitsname(self):
        """Return the name of the object as filetype fits-image."""
        typename = self.fullname + '.fits'
        return typename
#
#----------------------------------------------------------------------------
    
#----------------------------------------------------------------------------
#
#
class GrismStampName:
    """Class to derive filenames for the grism stamp image
    The filenames are composed bye three items:
    1. The specific rootname 'stpgri'
    2. A postfix to the rootname, mostly the object number
    3. An extension to derive filenames for various filetypes

    There are are only get and set methods to compose and get
    the full names for the various filetypes. Although
    there is a method to change the rootname, it is not recommended
    to do so since the rootname is supposed to be proprietary
    to the class.
    """
#----------------------------------------------------------------------------
#
    nameroot = 'stampgri'
#
#----------------------------------------------------------------------------

    def __init__(self, postfix):
        """The rootname is combined with the postfix at initialization.
        The postfix is stored in case that later a new rootname is given,
        and the a new name has to be built"""
        self.postfix = str(postfix)
        self.fullname = self.nameroot + str(postfix)

        self.defaultimname = self.getpngname()

    def getdefimname(self):
        return self.defaultimname
    def setname(self, postfix):
        """Set a new name by giving a new postfix."""
        self.postfix = str(postfix)
        self.fullname = self.nameroot + str(postfix)
    def getname(self):
        """Return the actual name of the 'object'."""
        return self.fullname
    def getpostfix(self):
        """Return the actual postfix of the 'object'."""
        return self.postfix
    def setrootname(self, rootname):
        """Set a new rootname (deprecated!!)."""
        self.nameroot = rootname
        self.fullname = self.nameroot + self.postfix
    def getrootname(self):
        """Return the root name of the object."""
        return self.nameroot
    def getjpgname(self):
        """Return the name of the object as filetype jpeg-image."""
        typename = self.fullname + '.jpg'
        return typename
    def getpngname(self):
        """Return the name of the object as filetype png-image."""
        typename = self.fullname + '.png'
        return typename
    def getpbmname(self):
        """Return the name of the object as filetype png-image."""
        typename = self.fullname + '.pbm'
        return typename
    def getgifname(self):
        """Return the name of the object as filetype gif-image."""
        typename = self.fullname + '.gif'
        return typename
    def gettifname(self):
        """Return the name of the object as filetype tif-image."""
        typename = self.fullname + '.tif'
        return typename
    def getfitsname(self):
        """Return the name of the object as filetype fits-image."""
        typename = self.fullname + '.fits'
        return typename
#
#----------------------------------------------------------------------------
    

#----------------------------------------------------------------------------
#
#
#
class ImageDirectory:
    """
    Small class to locate the image directory, that means
    the directory where the images sit which came along with the aXe2html
    distribution. This information is needed in case that images where not
    found and the default images are needed, or to access the logos.
    """
    def __init__(self):
        """
        Input:
            -

        Output:
            -

        Description:
            The main job is done at initialization:
            The path to the module 'aXe2html' is looked up in the appropriate
            dictionary, from there the image directory is located.
        """
        import string, sys
        import aXe2html
        axmod = str(sys.modules['aXe2html'])
        self.imagepath = axmod[string.find(axmod,'/'):string.rfind(axmod,'/')+1] + 'images/'
    def getimagedir(self):
        """
        Input:
            -

        Output:
            self.imagepath - the path to the image directory

        Description:
        Returns the full path to the image directrory for the
        current installation.
        """
        return self.imagepath
#
#----------------------------------------------------------------------------


#----------------------------------------------------------------------------
#
#
#
class DefaultImages:
    """
    Class for the administration of the default images for
    the direct stamp, grism image stamp and the plots.
    Those images are filled into the table in case that there
    are problems with the science images.
    """
    dirimage = 'nodimage'
    grismage = 'nogimage'
    ploimage = 'noplot'

    def __init__(self):
        """
        Takes as input only trhe image format. Then the instance
        variables are filled with the location of the default images
        in the correct image format.
        """
        self.imagedir = ImageDirectory().getimagedir()
        self.dimfull = self.imagedir + self.dirimage
        self.grifull = self.imagedir + self.grismage
        self.plofull = self.imagedir + self.ploimage

        self.defdimfull = self.addpngext(self.dimfull)
        self.defgrifull = self.addpngext(self.grifull)
        self.defplofull = self.addpngext(self.plofull)

    def getdefdimname(self):
        """
        Input:
            -

        Output:
            self.defdimfull - the name of the direct image default

        Description:
            Returns the name of the direct image default
        """
        return self.defdimfull
    def getdefgriname(self):
        """
        Input:
            -

        Output:
            self.defgrifull - the name of the grism image default

        Description:
            Returns the name of the grism image default
        """
        return self.defgrifull
    def getdefploname(self):
        """
        Input:
            -

        Output:
            self.defplofull - the name of the plot default

        Description:
            Returns the name of the plot default
        """
        return self.defplofull
    def addjpegext(self, inname):
        return inname + '.jpg'
    def addpngext(self, inname):
        return inname + '.png'
    def addgifext(self, inname):
        return inname + '.gif'
    def addtiffext(self, inname):
        return inname + '.tif'
#
#----------------------------------------------------------------------------

#----------------------------------------------------------------------------
#
#
#
class DecorationFiles:
    """
    Class to handle the naming and working with the logos for aXe
    and probably later other images who only should look nice.
    """
#    visimage = 'ACS_aXevis03.jpg'
#    speimage = 'ACS_aXe03.jpg'
#    cssfile  = 'axe_default.css'
    def __init__(self):
        import os, shutil

        self.visimage = 'ACS_aXevis03.jpg'
        self.speimage = 'ACS_aXe03.jpg'
        self.cssfile  = 'axe_default.css'
        self.favicon  = 'favicon.ico'

        self.imdir = ImageDirectory().getimagedir()
        
    def getvisimage(self, outpath, imgpath):
        """
        Input:
            outpath - the directory for the output

        Output:
            self.visimage - the name of the aXe2html logo

        Description:
            Looks whether the logo has just been copied to the output
            directory. Does that if it has not yet been done and gives
            the name of the logo back.            
        """
        import os, shutil
        destfile = outpath + '/' + imgpath + '/' + self.visimage
        if  not os.path.isfile(destfile):
#            shutil.copyfile(ImageDirectory().getimagedir() + self.visimage, destfile)
            shutil.copyfile(self.imdir + self.visimage, destfile)

        self.rel_path = imgpath + '/' + self.visimage

        return self.rel_path

    def getfavicon(self, outpath, imgpath):
        """
        Input:
            outpath - the directory for the output

        Output:
            self.favicon - the name of the aXe favicon

        Description:
            Looks whether the favicon has just been copied to the output
            directory. Does that if it has not yet been done and gives
            the name of the logo back.            
        """
        import os, shutil

        destfile = outpath + '/' + self.favicon

        if not os.path.isfile(destfile):
            shutil.copyfile(self.imdir + self.favicon, destfile)

        self.rel_path = imgpath + '/' + self.favicon

        return self.rel_path

    def getspeimage(self, outpath, imgpath):
        """
        Input:
            outpath - the directory for the output

        Output:
            self.visimage - the name of the general aXe logo

        Description:
            Looks whether the logo has just been copied to the output
            directory. Does that if it has not yet been done and gives
            the name of the logo back.            
        """
        import os, shutil

        destfile = outpath + '/' + imgpath + '/' + self.speimage

        if not os.path.isfile(destfile):
            shutil.copyfile(self.imdir + self.speimage, destfile)

        self.rel_path = imgpath + '/' + self.speimage

        return self.rel_path

    def getcssfile(self, outpath, infile):
        """
        Input:
            outpath - the directory for the output
            infile  - the user supplied css filename

        Output:
            self.cssfile - the css filename

        Description:
            The method copies either the user supplied
            css file or the standard css file to
            the proper location in the output directory
        """
        import os, shutil
        if infile:
            pos = infile.rfind('/')
            self.cssfile = infile[pos+1:]
            srcfile = infile
        else:
            srcfile = self.imdir + self.cssfile

        destfile = outpath + '/' + self.cssfile

        shutil.copyfile(srcfile, destfile)

        return self.cssfile

    def getcssname(self, outpath, infile):
        """
        Input:
            outpath - the directory for the output

        Output:
            self.cssfile - the css filename

        Description:
            The method returns the name of the css file
            used in the current setup.
        """
        import os, shutil

        if infile:
            pos = infile.rfind('/')
            self.cssfile = infile[pos+1:]

        return self.cssfile
#
#----------------------------------------------------------------------------

#----------------------------------------------------------------------------
#
#
class SpectDataName:
    """Class to derive filenames for asciidata written to disk.
    The filenames are composed bye four items:
    1. The specific rootname 'spectrum'
    2. A postfix to the rootname, mostly the object number
    3. An extension to derive filenames for various filetypes

    There are are only the get methods to compose and get
    the full names for the various filetypes.
    """
#----------------------------------------------------------------------------
#
    nameroot = 'spectrum'
#
#----------------------------------------------------------------------------

    def __init__(self, postfix, defformat):
        """The rootname is combined with the postfix at initialization.
        The postfix is stored in case that later a new rootname is given,
        and the a new name has to be built"""

        self.fullname = self.nameroot + str(postfix)

        if defformat == 'GZ':
            self.defdataname = self.fullname + self.getgz_ext()
        elif defformat == 'BZ2':
            self.defdataname = self.fullname + self.getbz2_ext()
        elif defformat == 'TXT':
            self.defdataname = self.fullname + self.gettxt_ext()

    def getdefdataname(self):
        """Return the full name."""
        return self.defdataname
    def getgz_ext(self):
        """Return the extension of the object as gzipped file."""
        return '.dat.gz'
    def getbz2_ext(self):
        """Return the extension of the object as bzipped2 file."""
        return '.dat.bz2'
    def gettxt_ext(self):
        """Return the extension of the object as text file."""
        return '.dat'
#
#----------------------------------------------------------------------------
